# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines the ModelAnalysis class."""

import warnings
from typing import Any, List, Optional

import numpy as np
import pandas as pd

from responsibleai.feature_metadata import FeatureMetadata
from responsibleai.managers.causal_manager import CausalManager
from responsibleai.managers.counterfactual_manager import CounterfactualManager
from responsibleai.managers.error_analysis_manager import ErrorAnalysisManager
from responsibleai.managers.explainer_manager import ExplainerManager
from responsibleai.rai_insights import RAIInsights


class ModelAnalysis(object):

    """Defines the top-level Model Analysis API.
    Use ModelAnalysis to analyze errors, explain the most important
    features, compute counterfactuals and run causal analysis in a
    single API.
    """

    def __init__(self, model: Any, train: pd.DataFrame, test: pd.DataFrame,
                 target_column: str, task_type: str,
                 categorical_features: Optional[List[str]] = None,
                 train_labels: Optional[np.ndarray] = None,
                 serializer: Optional[Any] = None,
                 maximum_rows_for_test: int = 5000,
                 feature_metadata: Optional[FeatureMetadata] = None):
        """Creates a ModelAnalysis object.

        :param model: The model to compute RAI insights for.
            A model that implements sklearn.predict or sklearn.predict_proba
            or function that accepts a 2d ndarray.
        :type model: object
        :param train: The training dataset including the label column.
        :type train: pandas.DataFrame
        :param test: The test dataset including the label column.
        :type test: pandas.DataFrame
        :param target_column: The name of the label column.
        :type target_column: str
        :param task_type: The task to run, can be `classification` or
            `regression`.
        :type task_type: str
        :param categorical_features: The categorical feature names.
        :type categorical_features: list[str]
        :param train_labels: The class labels in the training dataset
        :type train_labels: numpy.ndarray
        :param serializer: Picklable custom serializer with save and load
            methods defined for model that is not serializable. The save
            method returns a dictionary state and load method returns the
            model.
        :type serializer: object
        :param maximum_rows_for_test: Limit on size of test data
            (for performance reasons)
        :type maximum_rows_for_test: int
        :param feature_metadata: Feature metadata for the train/test
                                 dataset to identify different kinds
                                 of features in the dataset.
        :type feature_metadata: FeatureMetadata
        """
        warnings.warn(
            "MODULE-DEPRECATION-WARNING: ModelAnalysis in responsibleai "
            "package is deprecated. Please use RAIInsights instead.",
            DeprecationWarning)
        self.rai_insights = RAIInsights(
            model,
            train,
            test,
            target_column,
            task_type,
            categorical_features=categorical_features,
            classes=train_labels,
            serializer=serializer,
            maximum_rows_for_test=maximum_rows_for_test,
            feature_metadata=feature_metadata)
        self.model = self.rai_insights.model
        self.train = self.rai_insights.train
        self.test = self.rai_insights.test
        self.target_column = self.rai_insights.target_column
        self.task_type = self.rai_insights.task_type
        self.categorical_features = self.rai_insights.categorical_features
        self._feature_metadata = feature_metadata

    @property
    def causal(self) -> CausalManager:
        """Get the causal manager.
        :return: The causal manager.
        :rtype: CausalManager
        """
        return self.rai_insights.causal

    @property
    def counterfactual(self) -> CounterfactualManager:
        """Get the counterfactual manager.
        :return: The counterfactual manager.
        :rtype: CounterfactualManager
        """
        return self.rai_insights.counterfactual

    @property
    def error_analysis(self) -> ErrorAnalysisManager:
        """Get the error analysis manager.
        :return: The error analysis manager.
        :rtype: ErrorAnalysisManager
        """
        return self.rai_insights.error_analysis

    @property
    def explainer(self) -> ExplainerManager:
        """Get the explainer manager.
        :return: The explainer manager.
        :rtype: ExplainerManager
        """
        return self.rai_insights.explainer

    def compute(self):
        """Calls compute on each of the managers."""
        self.rai_insights.compute()

    def list(self):
        """List information about each of the managers.
        :return: Information about each of the managers.
        :rtype: dict
        """
        return self.rai_insights.list()

    def get(self):
        """List information about each of the managers.

        :return: Information about each of the managers.
        :rtype: dict
        """
        return self.rai_insights.get()

    def get_data(self):
        """Get all data as ModelAnalysisData object

        :return: Model Analysis Data
        :rtype: ModelAnalysisData
        """
        return self.rai_insights.get_data()

    def save(self, path):
        """Save the ModelAnalysis to the given path.
        :param path: The directory path to save the ModelAnalysis to.
        :type path: str
        """
        self.rai_insights.save(path)

    @staticmethod
    def load(path):
        """Load the ModelAnalysis from the given path.
        :param path: The directory path to load the ModelAnalysis from.
        :type path: str
        :return: The ModelAnalysis object after loading.
        :rtype: ModelAnalysis
        """
        # create the ModelAnalysis without any properties using the __new__
        # function, similar to pickle
        inst = ModelAnalysis.__new__(ModelAnalysis)
        inst.rai_insights = RAIInsights.load(path)
        inst.model = inst.rai_insights.model
        inst.train = inst.rai_insights.train
        inst.test = inst.rai_insights.test
        inst.target_column = inst.rai_insights.target_column
        inst.task_type = inst.rai_insights.task_type
        inst.categorical_features = inst.rai_insights.categorical_features
        inst._feature_metadata = inst.rai_insights._feature_metadata
        return inst



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

from typing import Any, Dict, List, Optional

from raiutils.exceptions import UserConfigValidationException

_DROPPED_FEATURE_PURPOSE = 'dropped feature'
_CATEGORICAL_FEATURE_PURPOSE = 'categorical feature'
_IDENTITY_FEATURE_PURPOSE = 'identity feature name'
_DATETIME_FEATURE_PURPOSE = 'datetime feature'
_TIME_SERIES_ID_FEATURE_PURPOSE = 'time series ID feature'


class FeatureMetadata:
    def __init__(self,
                 identity_feature_name: Optional[str] = None,
                 datetime_features: Optional[List[str]] = None,
                 categorical_features: Optional[List[str]] = None,
                 dropped_features: Optional[List[str]] = None,
                 time_series_id_features: Optional[List[str]] = None):
        """Placeholder class for feature metadata provided by the user.

        :param identity_feature_name: Name of the feature which helps to
            uniquely identify a row or instance in user input dataset.
        :type identity_feature_name: Optional[str]
        :param datetime_features: Names of datetime features in the user input
            dataset.
        :type datetime_features: Optional[List[str]]
        :param categorical_features: List of categorical features in the
            user input dataset.
        :type categorical_features: Optional[List[str]]
        :param dropped_features: List of features that were dropped by the
            the user during training of their model.
        :type dropped_features: Optional[List[str]]
        :param time_series_id_features: List of features that are used
            to uniquely identify a time series in the user input dataset.
            Time series ID features are only relevant for forecasting, i.e.,
            when the task is 'forecasting'. Specifying time series ID features
            for other tasks will result in validation errors.
        :type time_series_id_features: Optional[List[str]]
        """
        if identity_feature_name is not None and not isinstance(
                identity_feature_name, str):
            raise UserConfigValidationException(
                'identity_feature_name must be a string.')
        FeatureMetadata._validate_column_list(
            _DATETIME_FEATURE_PURPOSE, datetime_features)
        FeatureMetadata._validate_column_list(
            _CATEGORICAL_FEATURE_PURPOSE, categorical_features)
        FeatureMetadata._validate_column_list(
            _DROPPED_FEATURE_PURPOSE, dropped_features)
        FeatureMetadata._validate_column_list(
            _TIME_SERIES_ID_FEATURE_PURPOSE, time_series_id_features)

        self.identity_feature_name = identity_feature_name
        self.datetime_features = datetime_features
        self.categorical_features = categorical_features
        self.dropped_features = dropped_features
        self.time_series_id_features = time_series_id_features

    def validate_feature_metadata_with_user_features(
            self, feature_names: List[str]):
        """Validate the user-provided feature metadata.

        :param feature_names: list of features in the dataset.
        :type feature_names: List[str]
        """
        identity_feature = ([self.identity_feature_name]
                            if self.identity_feature_name else None)
        FeatureMetadata._validate_columns(
            _DROPPED_FEATURE_PURPOSE, self.dropped_features, feature_names)
        FeatureMetadata._validate_columns(
            _CATEGORICAL_FEATURE_PURPOSE,
            self.categorical_features, feature_names)
        FeatureMetadata._validate_columns(
            _IDENTITY_FEATURE_PURPOSE, identity_feature, feature_names)
        FeatureMetadata._validate_columns(
            _DATETIME_FEATURE_PURPOSE, self.datetime_features, feature_names)
        FeatureMetadata._validate_columns(
            _TIME_SERIES_ID_FEATURE_PURPOSE, self.time_series_id_features,
            feature_names)

    def to_dict(self) -> Dict[str, Any]:
        """Convert the feature metadata to a dictionary.

        :return: Dictionary of feature metadata.
        :rtype: Dict[str, Any]
        """
        return {
            'identity_feature_name': self.identity_feature_name,
            'datetime_features': self.datetime_features,
            'categorical_features': self.categorical_features,
            'dropped_features': self.dropped_features,
            'time_series_id_features': self.time_series_id_features
        }

    def __eq__(self, other_feature_metadata) -> bool:
        """Compare the feature metadata with another feature metadata.

        :param other_feature_metadata: Feature metadata to compare with.
        :type other_feature_metadata: FeatureMetadata
        :return: True if the feature metadata is equal to the other feature
                 metadata.
        :rtype: bool
        """
        return self.identity_feature_name == \
            other_feature_metadata.identity_feature_name and \
            self.datetime_features == \
            other_feature_metadata.datetime_features and \
            self.categorical_features == \
            other_feature_metadata.categorical_features and \
            self.dropped_features == \
            other_feature_metadata.dropped_features and \
            self.time_series_id_features == \
            other_feature_metadata.time_series_id_features

    @staticmethod
    def _validate_column_list(
            column_purpose: str,
            column_names: Any):
        """Ensure the provided column list is a list of strings.

        :param column_purpose: The purpose the column fulfills in the dataset.
        :type column_purpose: str
        :param column_names: Column names to be validate.
        :type column_names: Any
        """
        if column_names is None:
            return
        if column_names is not None and not isinstance(
                column_names, list):
            raise UserConfigValidationException(
                '{} must be a list.'.format(column_purpose))

        for column_name in column_names:
            if not isinstance(column_name, str):
                raise UserConfigValidationException(
                    'Entries in {} must be strings.'.format(column_purpose))

    @staticmethod
    def _validate_columns(
            column_purpose: str,
            column_names: List[str],
            feature_names: List[str]):
        """Ensure the provided column is present in the dataset.

        :param column_purpose: The purpose the column fulfills in the dataset.
        :type column_purpose: str
        :param column_names: List of column names to validate.
        :type column_names: List[str]
        :param feature_names: List of features in the input dataset.
        :type feature_names: List[str]
        """
        if column_names is None:
            return
        feature_name_set = set(feature_names)
        for column_name in column_names:
            if column_name not in feature_name_set:
                raise UserConfigValidationException(
                    f'The given {column_purpose} {column_name} is not present '
                    f'in the provided features: {", ".join(feature_names)}.')



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import numpy as np
import pandas as pd
import pytest
from tests.common_utils import create_iris_data

from erroranalysis._internal.constants import (ARG, COLUMN, METHOD, PRED_Y,
                                               ROW_INDEX, TRUE_Y)
from rai_test_utils.datasets.tabular import (create_housing_data,
                                             create_simple_titanic_data)
from rai_test_utils.models.sklearn import (
    create_sklearn_random_forest_regressor, create_sklearn_svm_classifier,
    create_titanic_pipeline)
from raiutils.cohort import CohortFilterMethods
from raiutils.models import ModelTask
from responsibleai.rai_insights import RAIInsights

TOL = 1e-10
SEPAL_WIDTH = 'sepal width'
EMBARKED = 'embarked'
CLASSIFICATION_OUTCOME = 'Classification outcome'
REGRESSION_ERROR = 'Regression error'


class TestCohortFilterRAIInsights(object):
    def test_cohort_filter_equal(self):
        X_train, X_test, y_train, y_test, feature_names, _ = \
            create_iris_pandas()
        filters = [{ARG: [2.8],
                    COLUMN: SEPAL_WIDTH,
                    METHOD: CohortFilterMethods.METHOD_EQUAL}]
        validation_data = create_validation_data(X_test, y_test)
        validation_data = validation_data.loc[X_test[SEPAL_WIDTH] == 2.8]
        model_task = ModelTask.CLASSIFICATION
        model = create_sklearn_svm_classifier(X_train, y_train)
        categorical_features = []
        run_rai_insights(validation_data,
                         model,
                         X_train,
                         y_train,
                         X_test,
                         y_test,
                         feature_names,
                         categorical_features,
                         model_task,
                         filters=filters)

    @pytest.mark.parametrize('use_str_labels', [True, False])
    @pytest.mark.parametrize('target_type', ['Predicted Y', 'True Y'])
    def test_cohort_filter_target(self, target_type, use_str_labels):
        if target_type == 'Predicted Y':
            pytest.skip("Skipping this test due to a bug condition "
                        "in Predicted Y cohort filtering")
        X_train, X_test, y_train, y_test, feature_names, classes = \
            create_iris_pandas(use_str_labels)
        filters = [{ARG: [2],
                    COLUMN: target_type,
                    METHOD: CohortFilterMethods.METHOD_INCLUDES}]
        validation_data = create_validation_data(X_test, y_test)
        if use_str_labels:
            validation_filter = y_test == classes[2]
        else:
            validation_filter = y_test == 2
        validation_data = validation_data.loc[validation_filter]
        model_task = ModelTask.CLASSIFICATION
        model = create_sklearn_svm_classifier(X_train, y_train)
        categorical_features = []
        run_rai_insights(validation_data,
                         model,
                         X_train,
                         y_train,
                         X_test,
                         y_test,
                         feature_names,
                         categorical_features,
                         model_task,
                         filters=filters)

    def test_cohort_filter_less(self):
        X_train, X_test, y_train, y_test, feature_names, _ = \
            create_iris_pandas()
        filters = [{ARG: [2.8],
                    COLUMN: SEPAL_WIDTH,
                    METHOD: CohortFilterMethods.METHOD_LESS}]
        validation_data = create_validation_data(X_test, y_test)
        validation_data = validation_data.loc[X_test[SEPAL_WIDTH] < 2.8]
        model_task = ModelTask.CLASSIFICATION
        model = create_sklearn_svm_classifier(X_train, y_train)
        categorical_features = []
        run_rai_insights(validation_data,
                         model,
                         X_train,
                         y_train,
                         X_test,
                         y_test,
                         feature_names,
                         categorical_features,
                         model_task,
                         filters=filters)

    def test_cohort_filter_less_and_equal(self):
        X_train, X_test, y_train, y_test, feature_names, _ = \
            create_iris_pandas()

        filters = [{ARG: [2.8],
                    COLUMN: SEPAL_WIDTH,
                    METHOD: CohortFilterMethods.METHOD_LESS_AND_EQUAL}]
        validation_data = create_validation_data(X_test, y_test)
        validation_data = validation_data.loc[X_test[SEPAL_WIDTH] <= 2.8]
        model_task = ModelTask.CLASSIFICATION
        model = create_sklearn_svm_classifier(X_train, y_train)
        categorical_features = []
        run_rai_insights(validation_data,
                         model,
                         X_train,
                         y_train,
                         X_test,
                         y_test,
                         feature_names,
                         categorical_features,
                         model_task,
                         filters=filters)

    def test_cohort_filter_greater(self):
        X_train, X_test, y_train, y_test, feature_names, _ = \
            create_iris_pandas()
        filters = [{ARG: [2.8],
                    COLUMN: SEPAL_WIDTH,
                    METHOD: CohortFilterMethods.METHOD_GREATER}]
        validation_data = create_validation_data(X_test, y_test)
        validation_data = validation_data.loc[X_test[SEPAL_WIDTH] > 2.8]
        model_task = ModelTask.CLASSIFICATION
        model = create_sklearn_svm_classifier(X_train, y_train)
        categorical_features = []
        run_rai_insights(validation_data,
                         model,
                         X_train,
                         y_train,
                         X_test,
                         y_test,
                         feature_names,
                         categorical_features,
                         model_task,
                         filters=filters)

    def test_cohort_filter_greater_and_equal(self):
        X_train, X_test, y_train, y_test, feature_names, _ = \
            create_iris_pandas()
        filters = [{ARG: [2.8],
                    COLUMN: SEPAL_WIDTH,
                    METHOD: CohortFilterMethods.METHOD_GREATER_AND_EQUAL}]
        validation_data = create_validation_data(X_test, y_test)
        validation_data = validation_data.loc[X_test[SEPAL_WIDTH] >= 2.8]
        model_task = ModelTask.CLASSIFICATION
        model = create_sklearn_svm_classifier(X_train, y_train)
        categorical_features = []
        run_rai_insights(validation_data,
                         model,
                         X_train,
                         y_train,
                         X_test,
                         y_test,
                         feature_names,
                         categorical_features,
                         model_task,
                         filters=filters)

    def test_cohort_filter_in_the_range_of(self):
        X_train, X_test, y_train, y_test, feature_names, _ = \
            create_iris_pandas()
        filters = [{ARG: [2.8, 3.4],
                    COLUMN: SEPAL_WIDTH,
                    METHOD: CohortFilterMethods.METHOD_RANGE}]
        validation_data = create_validation_data(X_test, y_test)
        validation_data = validation_data.loc[
            (X_test[SEPAL_WIDTH] <= 3.4) & (X_test[SEPAL_WIDTH] >= 2.8)]
        model_task = ModelTask.CLASSIFICATION
        model = create_sklearn_svm_classifier(X_train, y_train)
        categorical_features = []
        run_rai_insights(validation_data,
                         model,
                         X_train,
                         y_train,
                         X_test,
                         y_test,
                         feature_names,
                         categorical_features,
                         model_task,
                         filters=filters)

    @pytest.mark.parametrize('arg, correct_prediction',
                             [([1], False), ([0], True)])
    def test_cohort_filter_multiclass_classification_outcome(
            self, arg, correct_prediction):
        X_train, X_test, y_train, y_test, feature_names, _ = \
            create_iris_pandas()
        model = create_sklearn_svm_classifier(X_train, y_train)
        model_task = ModelTask.CLASSIFICATION
        categorical_features = []

        # the index 1, corresponds to incorrect prediction
        # the index 0 correspond to correct prediction
        filters = [{ARG: arg,
                    COLUMN: CLASSIFICATION_OUTCOME,
                    METHOD: CohortFilterMethods.METHOD_INCLUDES}]
        pred_y = model.predict(X_test)
        validation_data = create_validation_data(X_test, y_test, pred_y)
        if correct_prediction:
            validation_filter = validation_data[PRED_Y] == validation_data[
                TRUE_Y]
        else:
            validation_filter = validation_data[PRED_Y] != validation_data[
                TRUE_Y]
        validation_data = validation_data.loc[validation_filter]
        validation_data = validation_data.drop(columns=PRED_Y)
        model_task = ModelTask.CLASSIFICATION
        run_rai_insights(validation_data,
                         model,
                         X_train,
                         y_train,
                         X_test,
                         y_test,
                         feature_names,
                         categorical_features,
                         model_task,
                         filters=filters,
                         is_empty_validation_data=(not correct_prediction))

    def test_cohort_filter_includes(self):
        X_train, X_test, y_train, y_test, numeric, categorical = \
            create_simple_titanic_data()
        feature_names = categorical + numeric
        clf = create_titanic_pipeline(X_train, y_train)
        categorical_features = categorical
        # the indexes 0, 2 correspond to S, C
        filters = [{ARG: [0, 2],
                    COLUMN: EMBARKED,
                    METHOD: CohortFilterMethods.METHOD_INCLUDES}]
        validation_data = create_validation_data(X_test, y_test)
        filter_embarked = X_test[EMBARKED].isin(['S', 'C'])
        validation_data = validation_data.loc[filter_embarked]
        model_task = ModelTask.CLASSIFICATION
        run_rai_insights(validation_data,
                         clf,
                         X_train,
                         y_train,
                         X_test,
                         y_test,
                         feature_names,
                         categorical_features,
                         model_task,
                         filters=filters)

    def test_cohort_filter_excludes(self):
        X_train, X_test, y_train, y_test, numeric, categorical = \
            create_simple_titanic_data()
        feature_names = categorical + numeric
        clf = create_titanic_pipeline(X_train, y_train)
        categorical_features = categorical
        # the indexes other than 0, 2 correspond to Q
        filters = [{ARG: [0, 2],
                    COLUMN: EMBARKED,
                    METHOD: CohortFilterMethods.METHOD_EXCLUDES}]
        validation_data = create_validation_data(X_test, y_test)
        filter_embarked = X_test[EMBARKED].isin(['Q'])
        validation_data = validation_data.loc[filter_embarked]
        model_task = ModelTask.CLASSIFICATION
        run_rai_insights(validation_data,
                         clf,
                         X_train,
                         y_train,
                         X_test,
                         y_test,
                         feature_names,
                         categorical_features,
                         model_task,
                         filters=filters)

    @pytest.mark.parametrize('arg, outcome', [([1, 2], False), ([0, 3], True)])
    def test_cohort_filter_binary_classification_outcome(self, arg, outcome):
        X_train, X_test, y_train, y_test, numeric, categorical = \
            create_simple_titanic_data()
        feature_names = categorical + numeric
        clf = create_titanic_pipeline(X_train, y_train)
        categorical_features = categorical
        # the indexes 1, 2 correspond to false positives and false negatives
        # the indexes 0, 3 correspond to true positives and true negatives
        filters = [{ARG: arg,
                    COLUMN: CLASSIFICATION_OUTCOME,
                    METHOD: CohortFilterMethods.METHOD_INCLUDES}]
        pred_y = clf.predict(X_test)
        validation_data = create_validation_data(X_test, y_test, pred_y)
        if not outcome:
            validation_filter = validation_data[PRED_Y] != validation_data[
                TRUE_Y]
        else:
            validation_filter = validation_data[PRED_Y] == validation_data[
                TRUE_Y]
        validation_data = validation_data.loc[validation_filter]
        validation_data = validation_data.drop(columns=PRED_Y)
        model_task = ModelTask.CLASSIFICATION
        run_rai_insights(validation_data,
                         clf,
                         X_train,
                         y_train,
                         X_test,
                         y_test,
                         feature_names,
                         categorical_features,
                         model_task,
                         filters=filters)

    def test_cohort_filter_index(self):
        X_train, X_test, y_train, y_test, feature_names, _ = \
            create_iris_pandas()
        # filter on index, which can be done from the RAI dashboard
        filters = [{ARG: [40],
                    COLUMN: ROW_INDEX,
                    METHOD: CohortFilterMethods.METHOD_LESS}]
        validation_data = create_validation_data(X_test, y_test)
        validation_data = validation_data.loc[validation_data[ROW_INDEX] <= 40]
        model_task = ModelTask.CLASSIFICATION
        model = create_sklearn_svm_classifier(X_train, y_train)
        categorical_features = []
        model_task = ModelTask.CLASSIFICATION
        run_rai_insights(validation_data,
                         model,
                         X_train,
                         y_train,
                         X_test,
                         y_test,
                         feature_names,
                         categorical_features,
                         model_task,
                         filters=filters)

    def test_cohort_filter_regression_error(self):
        X_train, X_test, y_train, y_test, feature_names = \
            create_housing_data()
        X_train = pd.DataFrame(X_train, columns=feature_names)
        X_test = pd.DataFrame(X_test, columns=feature_names)

        # filter on regression error, which can be done from the
        # RAI dashboard
        filters = [{ARG: [40],
                    COLUMN: REGRESSION_ERROR,
                    METHOD: CohortFilterMethods.METHOD_LESS_AND_EQUAL}]

        model = create_sklearn_random_forest_regressor(X_train, y_train)
        pred_y = model.predict(X_test)

        validation_data = create_validation_data(X_test, y_test, pred_y)
        validation_filter = abs(validation_data[PRED_Y] - validation_data[
            TRUE_Y]) <= 40.0
        validation_data = validation_data.loc[validation_filter]
        validation_data = validation_data.drop(columns=PRED_Y)

        model_task = ModelTask.REGRESSION
        categorical_features = []
        run_rai_insights(validation_data,
                         model,
                         X_train,
                         y_train,
                         X_test,
                         y_test,
                         feature_names,
                         categorical_features,
                         model_task,
                         filters=filters)


def create_iris_pandas(use_str_labels=False):
    X_train, X_test, y_train, y_test, feature_names, classes = \
        create_iris_data()

    X_train = pd.DataFrame(X_train, columns=feature_names)
    X_test = pd.DataFrame(X_test, columns=feature_names)

    if use_str_labels:
        y_train = np.array([classes[y] for y in y_train])
        y_test = np.array([classes[y] for y in y_test])

    return X_train, X_test, y_train, y_test, feature_names, classes


def create_validation_data(X_test, y_test, pred_y=None):
    validation_data = X_test.copy()
    validation_data[TRUE_Y] = y_test
    validation_data[ROW_INDEX] = np.arange(0, len(y_test))
    if pred_y is not None:
        validation_data[PRED_Y] = pred_y
    return validation_data


def run_rai_insights(validation_data,
                     model,
                     X_train,
                     y_train,
                     X_test,
                     y_test,
                     feature_names,
                     categorical_features,
                     model_task,
                     filters=None,
                     composite_filters=None,
                     is_empty_validation_data=False):
    train = X_train.copy()
    train["target"] = y_train
    test = X_test.copy()
    test["target"] = y_test
    rai_insights = RAIInsights(
        model, train, test, "target", model_task,
        categorical_features=categorical_features)

    filtered_data = rai_insights.get_filtered_test_data(
        filters,
        composite_filters)

    # validate there is some data selected for each of the filters
    if is_empty_validation_data:
        assert validation_data.shape[0] == 0
    else:
        assert validation_data.shape[0] > 0
    assert validation_data.equals(filtered_data)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.


from responsibleai.exceptions import (ConfigAndResultMismatchException,
                                      DuplicateManagerConfigException,
                                      SchemaErrorException,
                                      SystemErrorException,
                                      UserConfigValidationException,
                                      UserErrorException)

TEST_MESSAGE = "Test message"


class TestExceptions:
    def _verify_exception_hierarchy(
            self, exception, parent_exception,
            error_code, exception_message):
        assert exception._error_code == error_code
        assert str(exception) == exception_message
        assert isinstance(exception, type(parent_exception))

    def test_user_exceptions(self):
        ucve = UserConfigValidationException(TEST_MESSAGE)
        self._verify_exception_hierarchy(
            ucve, UserErrorException(), 'Invalid config', TEST_MESSAGE)

        de = DuplicateManagerConfigException(TEST_MESSAGE)
        self._verify_exception_hierarchy(
            de, UserErrorException(),
            'Duplicate RAI configuration detected', TEST_MESSAGE)

        uee = UserErrorException(TEST_MESSAGE)
        self._verify_exception_hierarchy(
            uee, UserErrorException(),
            'User Error', TEST_MESSAGE)

    def test_system_exceptions(self):
        cerm = ConfigAndResultMismatchException(TEST_MESSAGE)
        self._verify_exception_hierarchy(
            cerm, SystemErrorException(),
            'Config and result mismatch', TEST_MESSAGE)

        see = SchemaErrorException(TEST_MESSAGE)
        self._verify_exception_hierarchy(
            see, SystemErrorException(),
            'Schema Error', TEST_MESSAGE)

        se = SystemErrorException(TEST_MESSAGE)
        self._verify_exception_hierarchy(
            se, SystemErrorException(),
            'System Error', TEST_MESSAGE)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import pandas as pd

from responsibleai.databalanceanalysis import AggregateBalanceMeasures

from ..common_utils import assert_series_and_dict_equal
from .conftest import SYNTHETIC_DATA_ETHNICITY, SYNTHETIC_DATA_GENDER


class TestAggregateBalanceMeasures:
    def test_one_feature_synthetic_data(
        self, synthetic_data, expected_aggregate_measures_gender
    ):
        agg_measures = (
            AggregateBalanceMeasures(cols_of_interest=[SYNTHETIC_DATA_GENDER])
            .measures(dataset=synthetic_data)
            .iloc[0]
        )
        assert_series_and_dict_equal(
            agg_measures, expected_aggregate_measures_gender
        )

    def test_both_features_synthetic_data(
        self, synthetic_data, expected_aggregate_measures_gender_ethnicity
    ):
        agg_measures = (
            AggregateBalanceMeasures(
                cols_of_interest=[
                    SYNTHETIC_DATA_GENDER,
                    SYNTHETIC_DATA_ETHNICITY,
                ]
            )
            .measures(dataset=synthetic_data)
            .iloc[0]
        )
        assert_series_and_dict_equal(
            agg_measures, expected_aggregate_measures_gender_ethnicity
        )

    def test_adult_data(
        self, adult_data, adult_data_aggregate_balance_measures
    ):
        train_df, test_df, _, _, _ = adult_data
        full_df = pd.concat([train_df, test_df])
        actual = AggregateBalanceMeasures(
            cols_of_interest=["race", "gender"]
        ).measures(dataset=full_df)
        pd.testing.assert_frame_equal(
            actual.reset_index(drop=True),
            adult_data_aggregate_balance_measures.reset_index(drop=True),
        )



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import json

import requests

from responsibleai.serialization_utilities import serialize_json_safe


class ServedModelWrapper:
    """Wrapper for locally served model.

    The purpose of ServedModelWrapper is to provide an abstraction
    for locally served models. This allows us to use the same code in
    RAIInsights for loaded models that can run in the same environment and
    also for locally served models.

    Locally served in this case means on localhost via HTTP.
    This could be in a separate conda environment, or even in a Docker
    container.

    :param port: The port on which the model is served.
    :type port: int
    """
    def __init__(self, port):
        self.port = port

    def forecast(self, X):
        """Get forecasts from the model.

        :param X: The input data.
        :type X: pandas.DataFrame
        :return: The model's forecasts based on the input data.
        :rtype: List[float]
        """
        # request formatting according to mlflow docs
        # https://mlflow.org/docs/latest/cli.html#mlflow-models-serve
        # JSON safe serialization takes care of datetime columns
        response = requests.post(
            url=f"http://localhost:{self.port}/invocations",
            headers={"Content-Type": "application/json"},
            data=json.dumps(
                {"dataframe_split": X.to_dict(orient='split')},
                default=serialize_json_safe))
        try:
            response.raise_for_status()
        except Exception:
            raise RuntimeError(
                "Could not retrieve predictions. "
                f"Model server returned status code {response.status_code} "
                f"and the following response: {response.content}")

        # json.loads decodes byte string response.
        # Response is a dictionary with a single entry 'predictions'
        return json.loads(response.content)['predictions']



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

# All wrappers in this file will be migrated to ml-wrappers.

import inspect
import sys

import pandas as pd
from ml_wrappers.model.base_wrapped_model import BaseWrappedModel

from raiutils.models import Forecasting, ModelTask

_AZUREML = "azureml"
_SKTIME = "sktime"


def _get_model_package(model):
    """Return the package name of the passed model."""
    module = inspect.getmodule(model)
    module_name = module.__name__.partition('.')[0]
    model_package = sys.modules[module_name].__package__
    return model_package


def _wrap_model(model, examples, time_feature, time_series_id_features):
    """Choose the appropriate wrapper for the model.

    If the model has a forecast_quantiles method
    (or in the sktime case: predict_quantiles) then choose the wrapped
    quantile forecasting model, otherwise the basic wrapper without
    quantile forecasting capabilities.
    """
    model_package = _get_model_package(model)
    if ((model_package == _SKTIME and hasattr(model, "predict_quantiles")) or
            hasattr(model, Forecasting.FORECAST_QUANTILES)):
        return _WrappedQuantileForecastingModel(
            model, examples, time_feature, time_series_id_features)
    else:
        return _WrappedForecastingModel(
            model, examples, time_feature, time_series_id_features)


class _WrappedForecastingModel(BaseWrappedModel):
    """Wrapper for forecasting models.

    This provides a unified API for the forecast method.
    Additionally, this wrapper knows how to work with models from
    certain packages such as AzureML and sktime.
    Other models are expected to have a forecast method that takes
    a DataFrame with the time series ID features and time feature.
    """
    def __init__(self, model, examples, time_feature,
                 time_series_id_features):
        super(_WrappedForecastingModel, self).__init__(
            model,
            eval_function=None,
            examples=examples,
            model_task=ModelTask.FORECASTING)
        self._model_package = _get_model_package(model)
        if (self._model_package not in [_AZUREML, _SKTIME] and
                not hasattr(model, Forecasting.FORECAST)):
            raise ValueError(
                "The passed model does not have a 'forecast' method. "
                "'forecast' is required for the forecasting task_type. "
                "Alternatively, pass a sktime forecasting model.")
        self._validate_time_features(time_feature, time_series_id_features)

    def _validate_time_features(self, time_feature, time_series_id_features):
        """Ensures that time and time series ID features are present."""
        if time_feature not in self._examples.columns:
            raise ValueError(
                "The passed time_feature is not in the examples DataFrame.")
        self._time_feature = time_feature
        for ts_id in time_series_id_features:
            if ts_id not in self._examples.columns:
                raise ValueError(
                    f"The passed time series ID feature {ts_id} is not "
                    "in the examples DataFrame.")
        self._time_feature = time_feature
        self._time_series_id_features = time_series_id_features

    def forecast(self, X):
        """Returns the forecast for the passed data."""
        if self._model_package == _AZUREML:
            # AzureML forecasting models return a tuple of (forecast, data)
            # but we only want to return the actual forecast.
            return self._model.forecast(X)[0]
        if self._model_package == _SKTIME:
            return self._apply_sktime_method(self._model.predict, X)

        # default case
        return self._model.forecast(X)

    def _get_forecast_horizon(self, X):
        """Returns the forecast horizon for the passed data."""
        # This method is only called if the model is from the sktime package
        # so we can assume that sktime is installed.
        # For non-sktime or more generally non-forecasting scenarios sktime
        # won't be needed so we don't want to add it as a requirement to the
        # responsibleai package.
        from sktime.forecasting.base import ForecastingHorizon
        return ForecastingHorizon(
            pd.to_datetime(X[self._time_feature].unique()).sort_values(),
            is_relative=False)

    def _apply_sktime_method(self, method, X):
        """Applies the passed sktime forecasting method.

        This method is just a helper for sktime models.
        """
        fh = self._get_forecast_horizon(X)
        # sktime expects the time series ID features and time feature
        # in the index rather than as a regular column.
        X_temp = X.copy()
        X_temp.set_index(
            self._time_series_id_features + [self._time_feature],
            inplace=True, drop=True)
        if len(self._time_series_id_features) == 0:
            return method(X=X_temp, fh=fh)

        # If there are potentially multiple time series in the data
        # we need to ensure that sktime receives data for all of them.
        # This is currently an issue in sktime:
        # https://github.com/sktime/sktime/issues/4209
        # When this is supported in sktime we can remove the code in this
        # if-branch.

        # Check that all input time series have all required rows.
        time_series_id_features_index_levels = \
            list(range(len(self._time_series_id_features)))
        time_series_counts = X_temp.groupby(
            level=time_series_id_features_index_levels).size().to_list()
        all_time_series_have_same_number_of_rows = \
            all(count == time_series_counts[0] for count in time_series_counts)
        if not all_time_series_have_same_number_of_rows:
            raise ValueError(
                "Not all time series have the same number of rows.")
        # Determine which time series are missing from the data.
        # All index levels except for the last one are time series ID features.
        # The last level is the datetime feature.
        existing_time_series = \
            X_temp.index.droplevel(level=-1).unique().to_list()
        all_time_series = self._model.forecasters_.index.to_list()
        missing_time_series = \
            list(set(all_time_series) - set(existing_time_series))
        # Add the missing time series to the data by duplicating
        # the first time series.
        n_rows = int(len(X_temp) / len(existing_time_series))
        for time_series in missing_time_series:
            X_add = X_temp.iloc[:n_rows].copy()
            id_feature_value_mapping = \
                dict(zip(self._time_series_id_features, time_series))
            for id_feature, value in id_feature_value_mapping.items():
                X_add[id_feature] = value
            X_add[self._time_feature] = X_add.index.get_level_values(level=-1)
            X_add.set_index(
                self._time_series_id_features + [self._time_feature],
                drop=True, inplace=True)
            X_temp = pd.concat((X_temp, X_add))

        preds = method(X=X_temp, fh=fh)
        # Remove the predictions that weren't requested.
        return preds.head(len(X))


class _WrappedQuantileForecastingModel(_WrappedForecastingModel):
    """Wrapper for quantile forecasting models.

    This provides a unified API for the forecast_quantile method
    in addition to the forecast method provided by the base class.
    Additionally, this wrapper knows how to work with models from
    certain packages such as AzureML and sktime.
    Other models are expected to have a forecast method that takes
    a DataFrame with the time series ID features and time feature.
    """
    def __init__(self, model, examples, time_feature,
                 time_series_id_features):
        super(_WrappedQuantileForecastingModel, self).__init__(
            model,
            examples=examples,
            time_feature=time_feature,
            time_series_id_features=time_series_id_features)
        if (self._model_package not in [_AZUREML, _SKTIME] and
                not hasattr(model, Forecasting.FORECAST_QUANTILES)):
            raise ValueError(
                "The passed model does not have a 'forecast_quantiles' "
                "method. 'forecast_quantiles' is optional for the "
                "forecasting task_type, but required for "
                "WrappedQuantileForecastingModel. Alternatively, pass a "
                "sktime forecasting model that supports predict_quantiles.")

    def forecast_quantiles(self, X, quantiles=None):
        """Returns the forecast for the given quantiles.

        If no quantiles are passed, the default quantiles 0.025 and 0.975
        are used.
        """
        if quantiles is None:
            quantiles = [0.025, 0.975]
        if (not isinstance(quantiles, list) or
                len(quantiles) == 0 or
                any([not isinstance(q, float) or
                     q <= 0 or
                     q >= 1 for q in quantiles])):
            raise ValueError(
                "quantiles must be a list of floats between 0 and 1.")
        if self._model_package == _AZUREML:
            # AzureML forecasting models don't take quantiles as an
            # argument but instead need to have it set on the model
            # object.
            self._model.quantiles = quantiles
            return self._model.forecast_quantiles(X)
        if self._model_package == _SKTIME:
            def forecast_quantiles_method(X, fh):
                return self._model.predict_quantiles(
                    X=X, fh=fh, alpha=quantiles)
            return self._apply_sktime_method(forecast_quantiles_method, X)
        return self._model.forecast_quantiles(X, quantiles)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import os
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd
import pytest
from tests.common_utils import (ADULT_CATEGORICAL_FEATURES_AFTER_DROP,
                                ADULT_DROPPED_FEATURES, create_iris_data)

from rai_test_utils.datasets.tabular import \
    create_binary_classification_dataset
from rai_test_utils.models.lightgbm import create_lightgbm_classifier
from rai_test_utils.models.sklearn import \
    create_complex_classification_pipeline
from raiutils.models import ModelTask
from responsibleai import RAIInsights
from responsibleai._internal.constants import (ManagerNames,
                                               SerializationAttributes)
from responsibleai.feature_metadata import FeatureMetadata

LABELS = 'labels'


class TestRAIInsightsSaveAndLoadScenarios(object):

    def test_rai_insights_empty_save_load_save(self):
        X_train, y_train, X_test, y_test, classes = \
            create_binary_classification_dataset()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train[LABELS] = y_train
        X_test[LABELS] = y_test

        rai_insights = RAIInsights(
            model, X_train, X_test,
            LABELS,
            categorical_features=None,
            task_type=ModelTask.CLASSIFICATION)

        with TemporaryDirectory() as tmpdir:
            save_1 = Path(tmpdir) / "first_save"
            save_2 = Path(tmpdir) / "second_save"

            # Save it
            rai_insights.save(save_1)
            assert len(os.listdir(save_1 / ManagerNames.CAUSAL)) == 0
            assert len(os.listdir(save_1 / ManagerNames.COUNTERFACTUAL)) == 0
            assert len(os.listdir(save_1 / ManagerNames.DATA_BALANCE)) == 0
            assert len(os.listdir(save_1 / ManagerNames.ERROR_ANALYSIS)) == 0
            assert len(os.listdir(save_1 / ManagerNames.EXPLAINER)) == 0

            # Load
            rai_2 = RAIInsights.load(save_1)

            # Validate, but this isn't the main check
            validate_rai_insights(
                rai_2, X_train, X_test,
                LABELS, ModelTask.CLASSIFICATION)

            # Save again (this is where Issue #1046 manifested)
            rai_2.save(save_2)
            assert len(os.listdir(save_2 / ManagerNames.CAUSAL)) == 0
            assert len(os.listdir(save_2 / ManagerNames.COUNTERFACTUAL)) == 0
            assert len(os.listdir(save_2 / ManagerNames.DATA_BALANCE)) == 0
            assert len(os.listdir(save_2 / ManagerNames.ERROR_ANALYSIS)) == 0
            assert len(os.listdir(save_2 / ManagerNames.EXPLAINER)) == 0

    @pytest.mark.parametrize('manager_type', [ManagerNames.CAUSAL,
                                              ManagerNames.ERROR_ANALYSIS,
                                              ManagerNames.EXPLAINER,
                                              ManagerNames.COUNTERFACTUAL,
                                              ManagerNames.DATA_BALANCE])
    def test_rai_insights_save_load_add_save(self, manager_type, adult_data):
        data_train, data_test, y_train, y_test, categorical_features, \
            continuous_features, target_name, classes, \
            feature_columns, feature_range_keys = adult_data
        X_train = data_train.drop([target_name], axis=1)

        model = create_complex_classification_pipeline(
            X_train, y_train, continuous_features, categorical_features)

        # Cut down size for counterfactuals, in the interests of speed
        if manager_type == ManagerNames.COUNTERFACTUAL:
            data_test = data_test[0:1]

        feature_metadata = FeatureMetadata(identity_feature_name="age")
        save_load_add_rai_insights(
            manager_type=manager_type,
            data_train=data_train,
            data_test=data_test,
            target_name=target_name,
            categorical_features=categorical_features,
            feature_columns=feature_columns,
            feature_range_keys=feature_range_keys,
            feature_metadata=feature_metadata,
            model=model)

    @pytest.mark.parametrize('target_dir', [ManagerNames.CAUSAL,
                                            ManagerNames.ERROR_ANALYSIS,
                                            ManagerNames.COUNTERFACTUAL])
    def test_load_missing_dirs(self, target_dir, adult_data):
        # This test is about the case where an object has been saved to Azure
        # Directories only exist implicitly, so in a downloaded instance
        # if a manager had no outputs, then its subdirectory won't exist
        # The exception is the Explainer, which always creates a file
        # in its subdirectory
        data_train, data_test, y_train, y_test, categorical_features, \
            continuous_features, target_name, classes, \
            feature_columns, feature_range_keys = adult_data
        X_train = data_train.drop([target_name], axis=1)

        model = create_complex_classification_pipeline(
            X_train, y_train, continuous_features, categorical_features)
        rai_insights = RAIInsights(
            model, data_train, data_test,
            target_name,
            categorical_features=categorical_features,
            task_type=ModelTask.CLASSIFICATION)

        with TemporaryDirectory() as tmpdir:
            save_1 = Path(tmpdir) / "first_save"

            # Save it
            rai_insights.save(save_1)

            # Remove the target directory
            # First make sure it's empty
            dir_to_remove = save_1 / target_dir
            assert len(list(dir_to_remove.iterdir())) == 0
            os.rmdir(dir_to_remove)
            assert not dir_to_remove.exists()

            # Load
            rai_2 = RAIInsights.load(save_1)
            assert rai_2 is not None

    def test_loading_rai_insights_without_model_file(self):
        X_train, X_test, y_train, y_test, feature_names, classes = \
            create_iris_data()
        model = create_lightgbm_classifier(X_train, y_train)
        X_train['target'] = y_train
        X_test['target'] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test,
            target_column='target',
            task_type='classification')

        with TemporaryDirectory() as tmpdir:
            assert rai_insights.model is not None
            save_path = Path(tmpdir) / "rai_insights"
            rai_insights.save(save_path)

            # Remove the model.pkl file to cause an exception to occur
            # while loading the model.
            model_pkl_path = Path(tmpdir) / \
                "rai_insights" / SerializationAttributes.MODEL_PKL
            os.remove(model_pkl_path)
            with pytest.raises(Exception):
                RAIInsights.load(save_path)

    @pytest.mark.parametrize('manager_type', [ManagerNames.CAUSAL,
                                              ManagerNames.ERROR_ANALYSIS,
                                              ManagerNames.EXPLAINER,
                                              ManagerNames.COUNTERFACTUAL,
                                              ManagerNames.DATA_BALANCE])
    def test_rai_insights_add_save_load_save(self, manager_type, adult_data):
        data_train, data_test, y_train, y_test, categorical_features, \
            continuous_features, target_name, classes, \
            feature_columns, feature_range_keys = adult_data
        X_train = data_train.drop([target_name], axis=1)

        model = create_complex_classification_pipeline(
            X_train, y_train, continuous_features, categorical_features)

        # Cut down size for counterfactuals, in the interests of speed
        if manager_type == ManagerNames.COUNTERFACTUAL:
            data_test = data_test[0:1]

        rai_insights = RAIInsights(
            model, data_train, data_test,
            target_name,
            categorical_features=categorical_features,
            task_type=ModelTask.CLASSIFICATION,
            feature_metadata=FeatureMetadata(identity_feature_name="age"))

        # Call a single manager
        if manager_type == ManagerNames.CAUSAL:
            rai_insights.causal.add(
                treatment_features=['age', 'hours_per_week']
            )
        elif manager_type == ManagerNames.COUNTERFACTUAL:
            rai_insights.counterfactual.add(
                total_CFs=10,
                desired_class='opposite',
                feature_importance=False
            )
        elif manager_type == ManagerNames.DATA_BALANCE:
            rai_insights._data_balance_manager.add(
                cols_of_interest=categorical_features
            )
        elif manager_type == ManagerNames.ERROR_ANALYSIS:
            rai_insights.error_analysis.add()
        else:
            rai_insights.explainer.add()

        rai_insights.compute()

        with TemporaryDirectory() as tmpdir:
            save_1 = Path(tmpdir) / "first_save"
            save_2 = Path(tmpdir) / "second_save"

            # Save it
            rai_insights.save(save_1)

            # Load
            rai_2 = RAIInsights.load(save_1)

            # Validate, but this isn't the main check
            validate_rai_insights(
                rai_2, data_train, data_test,
                target_name, ModelTask.CLASSIFICATION,
                categorical_features=categorical_features,
                feature_range_keys=feature_range_keys,
                feature_columns=feature_columns,
                feature_metadata=FeatureMetadata(identity_feature_name="age"))

            # Save again (this is where Issue #1081 manifested)
            rai_2.save(save_2)

    @pytest.mark.parametrize('manager_type', [ManagerNames.CAUSAL,
                                              ManagerNames.ERROR_ANALYSIS,
                                              ManagerNames.EXPLAINER,
                                              ManagerNames.COUNTERFACTUAL,
                                              ManagerNames.DATA_BALANCE])
    def test_rai_insights_save_load_add_dropped_features(self, manager_type,
                                                         adult_data):
        data_train, data_test, y_train, y_test, categorical_features, \
            continuous_features, target_name, classes, \
            feature_columns, feature_range_keys = adult_data

        dropped_features = ADULT_DROPPED_FEATURES
        categorical_features_after_drop = \
            ADULT_CATEGORICAL_FEATURES_AFTER_DROP

        X_train = data_train.drop([target_name] + dropped_features, axis=1)

        model = create_complex_classification_pipeline(
            X_train, y_train, continuous_features,
            categorical_features_after_drop)

        # Cut down size for counterfactuals, in the interests of speed
        if manager_type == ManagerNames.COUNTERFACTUAL:
            data_test = data_test[0:1]

        feature_metadata = FeatureMetadata(dropped_features=dropped_features)

        save_load_add_rai_insights(
            manager_type=manager_type,
            data_train=data_train,
            data_test=data_test,
            target_name=target_name,
            categorical_features=categorical_features,
            feature_columns=feature_columns,
            feature_range_keys=feature_range_keys,
            feature_metadata=feature_metadata,
            model=model)


def save_load_add_rai_insights(
    manager_type,
    data_train,
    data_test,
    target_name,
    categorical_features,
    feature_columns,
    feature_range_keys,
    feature_metadata,
    model
):
    rai_insights = RAIInsights(
        model, data_train, data_test,
        target_name,
        categorical_features=categorical_features,
        task_type=ModelTask.CLASSIFICATION,
        feature_metadata=feature_metadata)

    with TemporaryDirectory() as tmpdir:
        save_1 = Path(tmpdir) / "first_save"
        save_2 = Path(tmpdir) / "second_save"

        # Save it
        rai_insights.save(save_1)

        # Load
        rai_2 = RAIInsights.load(save_1)

        # Call a single manager
        if manager_type == ManagerNames.CAUSAL:
            rai_2.causal.add(
                treatment_features=['age', 'hours_per_week']
            )
        elif manager_type == ManagerNames.COUNTERFACTUAL:
            rai_2.counterfactual.add(
                total_CFs=10,
                desired_class='opposite',
                feature_importance=False
            )
        elif manager_type == ManagerNames.DATA_BALANCE:
            rai_2._data_balance_manager.add(
                cols_of_interest=categorical_features
            )
        elif manager_type == ManagerNames.ERROR_ANALYSIS:
            rai_2.error_analysis.add()
        elif manager_type == ManagerNames.EXPLAINER:
            rai_2.explainer.add()
        else:
            raise ValueError(
                "Bad manager_type: {0}".format(manager_type))

        rai_2.compute()

        # Validate, but this isn't the main check
        validate_rai_insights(
            rai_2, data_train, data_test,
            target_name, ModelTask.CLASSIFICATION,
            categorical_features=categorical_features,
            feature_range_keys=feature_range_keys,
            feature_columns=feature_columns,
            feature_metadata=feature_metadata)

        # Save again (this is where Issue #1046 manifested)
        rai_2.save(save_2)


def validate_rai_insights(
    rai_insights,
    train_data,
    test_data,
    target_column,
    task_type,
    categorical_features=None,
    feature_range_keys=None,
    feature_columns=None,
    feature_metadata=None
):
    pd.testing.assert_frame_equal(rai_insights.train, train_data)
    pd.testing.assert_frame_equal(rai_insights.test, test_data)
    assert rai_insights.target_column == target_column
    assert rai_insights.task_type == task_type
    assert rai_insights.categorical_features == (categorical_features or [])
    if feature_range_keys is not None:
        assert feature_range_keys.sort() == \
            list(rai_insights._feature_ranges[0].keys()).sort()
    if feature_columns is not None:
        assert rai_insights._feature_columns == (feature_columns or [])
    if feature_metadata is not None:
        # mismatch between categorical_features passed in RAIInsights
        # constructor and the categorical_features set on the feature_metadata
        if (categorical_features is not None and
                feature_metadata.categorical_features is None):
            feature_metadata.categorical_features = categorical_features
        assert rai_insights._feature_metadata == feature_metadata
    assert target_column not in rai_insights._feature_columns

    if rai_insights.model is None:
        assert rai_insights._predict_output is None
        assert rai_insights._predict_proba_output is None
    else:
        assert rai_insights._predict_output is not None
        if task_type == ModelTask.CLASSIFICATION:
            assert rai_insights._predict_proba_output is not None
            assert isinstance(rai_insights._predict_proba_output, np.ndarray)
            assert len(rai_insights._predict_proba_output.tolist()[0]) == \
                len(rai_insights._classes)

    if task_type == ModelTask.CLASSIFICATION:
        classes = train_data[target_column].unique()
        classes.sort()
        np.testing.assert_array_equal(rai_insights._classes,
                                      classes)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.
import os
from unittest.mock import ANY, patch

import numpy as np
import pytest

from raiutils.exceptions import UserConfigValidationException
from raiutils.models import ModelTask
from responsibleai import RAIInsights
from responsibleai._interfaces import CausalData
from responsibleai._internal.constants import FileFormats
from responsibleai._tools.shared.state_directory_management import \
    DirectoryManager
from responsibleai.managers.causal_manager import CausalManager


class TestCausalManager:
    def test_causal_no_categoricals(self, housing_data):
        train_df, test_df, target_feature = housing_data

        manager = CausalManager(train_df, test_df, target_feature,
                                ModelTask.REGRESSION, None)
        manager.add(['AveRooms'])
        manager.compute()
        result = manager.get()[0]

        assert len(result.policies) == 1
        assert len(result.config.treatment_features) == 1
        assert result.config.treatment_features[0] == 'AveRooms'

    def test_causal_save_and_load(self, housing_data, tmpdir):
        train_df, test_df, target_feature = housing_data

        save_dir = tmpdir.mkdir('save-dir')

        insights = RAIInsights(
            None, train_df, test_df, target_feature, ModelTask.REGRESSION)

        insights.causal.add(['AveRooms'])
        insights.compute()

        pre_results = insights.causal.get()
        pre_result = pre_results[0]

        insights.causal._save(save_dir)
        manager = insights.causal._load(save_dir, insights)
        post_results = manager.get()
        post_result = post_results[0]
        assert post_result.id == pre_result.id
        assert post_result.causal_analysis is not None
        assert post_result.global_effects is not None
        assert post_result.local_effects is not None
        assert post_result.policies is not None

        # Remove the causal analysis models to test the loading of
        # causal models in case there is error in loading of the causal
        # models.
        all_causal_dirs = DirectoryManager.list_sub_directories(save_dir)
        for causal_dir in all_causal_dirs:
            dm = DirectoryManager(parent_directory_path=save_dir,
                                  sub_directory_name=causal_dir)
            causal_analysis_pkl_file_path = \
                dm.get_data_directory() / ("causal_analysis" + FileFormats.PKL)
            os.remove(causal_analysis_pkl_file_path)

        model_load_err = ('ERROR-LOADING-EXPLAINER: '
                          'There was an error loading the explainer. '
                          'Some of RAI dashboard features may not work.')
        with pytest.warns(UserWarning, match=model_load_err):
            manager = insights.causal._load(save_dir, insights)
        post_results = manager.get()
        post_result = post_results[0]
        assert post_result.id == pre_result.id
        assert post_result.causal_analysis is None
        assert post_result.global_effects is not None
        assert post_result.local_effects is not None
        assert post_result.policies is not None

    def test_causal_cat_expansion(self, parks_data):
        train_df, test_df, target_feature = parks_data

        manager = CausalManager(train_df, test_df, target_feature,
                                ModelTask.REGRESSION, ['state', 'attraction'])

        expected = "Increase the value 50"
        with pytest.raises(
                UserConfigValidationException, match=expected):
            manager.add(['state'])
            manager.compute()

    def test_causal_train_test_categories(self, parks_data):
        train_df, test_df, target_feature = parks_data

        test_df = test_df.copy()
        test_df.loc[len(test_df.index)] = ['indiana', 301, 'trees', 78]
        test_df.loc[len(test_df.index)] = ['indiana', 222, 'trees', 81]

        manager = CausalManager(train_df, test_df, target_feature,
                                ModelTask.REGRESSION, ['state', 'attraction'])

        message = ("Causal analysis requires that every category of "
                   "categorical features present in the test data be "
                   "also present in the train data. "
                   "Categories missing from train data: "
                   "{'state': \\['indiana'\\]}")
        with pytest.raises(UserConfigValidationException, match=message):
            manager.add(['state'],
                        skip_cat_limit_checks=True,
                        upper_bound_on_cat_expansion=50)

    def verify_common_causal_data_attributes(self, causal_data):
        assert isinstance(causal_data, CausalData)
        assert hasattr(causal_data, 'id')
        assert hasattr(causal_data, 'version')
        assert hasattr(causal_data, 'config')

    def test_causal_manager_global_cohort_effects(self, housing_data):
        train_df, test_df, target_feature = housing_data

        manager = CausalManager(train_df, test_df, target_feature,
                                ModelTask.REGRESSION, None)
        manager.add(['AveRooms'])
        manager.compute()

        id = manager.get()[0].id
        X_test = test_df.drop(target_feature, axis=1)
        causal_data = manager.request_global_cohort_effects(id, X_test)

        self.verify_common_causal_data_attributes(causal_data)
        assert hasattr(causal_data, 'global_effects')
        EFFECTS_ATTRIBUTES = [
            'point',
            'outcome',
            'feature',
            'feature_value',
            'stderr',
            'zstat',
            'ci_lower',
            'ci_upper',
            'p_value',
        ]
        for effect in EFFECTS_ATTRIBUTES:
            assert effect in causal_data.global_effects[0]

        incorrect_query_id = "incorrect_query_id"
        X_test = test_df.drop(target_feature, axis=1)
        with pytest.raises(
                ValueError,
                match="Failed to find causal result with ID: "
                      "incorrect_query_id"):
            manager.request_global_cohort_effects(
                incorrect_query_id, X_test)

    def test_causal_manager_global_cohort_policy(self, housing_data):
        train_df, test_df, target_feature = housing_data

        manager = CausalManager(train_df, test_df, target_feature,
                                ModelTask.REGRESSION, None)
        manager.add(['AveRooms'])
        manager.compute()

        id = manager.get()[0].id
        X_test = test_df.head(5).drop(target_feature, axis=1)
        causal_data = manager.request_global_cohort_policy(id, X_test)

        self.verify_common_causal_data_attributes(causal_data)
        assert hasattr(causal_data, 'policies')
        assert len(causal_data.policies[0].local_policies) == X_test.shape[0]
        assert causal_data.policies[0].treatment_feature == "AveRooms"

        incorrect_query_id = "incorrect_query_id"
        X_test = test_df.drop(target_feature, axis=1)
        with pytest.raises(
                ValueError,
                match="Failed to find causal result with ID: "
                      "incorrect_query_id"):
            manager.request_global_cohort_effects(
                incorrect_query_id, X_test)

    def test_causal_manager_local_instance_effects(self, housing_data):
        train_df, test_df, target_feature = housing_data

        manager = CausalManager(train_df, test_df, target_feature,
                                ModelTask.REGRESSION, None)
        manager.add(['AveRooms'])
        manager.compute()

        id = manager.get()[0].id
        X_test = test_df.head(1).drop(target_feature, axis=1)
        causal_data = manager.request_local_instance_effects(id, X_test)

        self.verify_common_causal_data_attributes(causal_data)
        assert hasattr(causal_data, 'local_effects')
        EFFECTS_ATTRIBUTES = [
            'sample',
            'outcome',
            'feature',
            'feature_value',
            'point',
            'stderr',
            'zstat',
            'ci_lower',
            'ci_upper',
            'p_value',
        ]
        for effect in EFFECTS_ATTRIBUTES:
            assert effect in causal_data.local_effects[0][0]

        incorrect_query_id = "incorrect_query_id"
        X_test = test_df.drop(target_feature, axis=1)
        with pytest.raises(
                ValueError,
                match="Failed to find causal result with ID: "
                      "incorrect_query_id"):
            manager.request_local_instance_effects(
                incorrect_query_id, X_test)

        id = manager.get()[0].id
        X_test = test_df.head(1).drop(target_feature, axis=1).values
        with pytest.raises(
                UserConfigValidationException,
                match='Data is of type <class \'numpy.ndarray\'>'
                      ' but it must be a pandas DataFrame.'):
            manager.request_local_instance_effects(id, X_test)

        id = manager.get()[0].id
        X_test = test_df.head(5).drop(target_feature, axis=1)
        with pytest.raises(
                UserConfigValidationException,
                match='Only one row of data is allowed for '
                      'local causal effects.'):
            manager.request_local_instance_effects(id, X_test)


@pytest.fixture(scope='class')
def cost_manager(housing_data):
    train_df, test_df, target_feature = housing_data

    test_df = test_df[:7]
    return CausalManager(train_df, test_df, target_feature,
                         ModelTask.REGRESSION, None)


class TestCausalManagerTreatmentCosts:
    def test_zero_cost(self, cost_manager):
        with patch.object(cost_manager, '_create_policy', return_value=None)\
                as mock_create:
            try:
                cost_manager.add(['AveRooms', 'Population', 'AveOccup'],
                                 treatment_cost=0)
                cost_manager.compute()
            except TypeError:
                pass
            mock_create.assert_any_call(ANY, ANY, 'AveRooms', 0, ANY, ANY, ANY)
            mock_create.assert_any_call(ANY, ANY, 'Population', 0,
                                        ANY, ANY, ANY)
            mock_create.assert_any_call(ANY, ANY, 'AveOccup', 0, ANY, ANY, ANY)

    def test_nonzero_scalar_cost(self, cost_manager):
        message = ("treatment_cost must be a list with the same number "
                   "of elements as treatment_features where each "
                   "element is either a constant cost of treatment "
                   "or an array specifying the cost of treatment per "
                   "sample. Found treatment_cost "
                   "of type <class 'int'>, expected list.")
        with pytest.raises(UserConfigValidationException, match=message):
            cost_manager.add(['AveRooms', 'Population', 'AveOccup'],
                             treatment_cost=5)

    def test_nonlist_cost(self, cost_manager):
        message = ("treatment_cost must be a list with the same number of "
                   "elements as treatment_features where each element is "
                   "either a constant cost of treatment or an array "
                   "specifying the cost of treatment per sample. "
                   "Found treatment_cost of type <class 'numpy.ndarray'>, "
                   "expected list.")
        with pytest.raises(UserConfigValidationException, match=message):
            cost_manager.add(['AveRooms', 'Population', 'AveOccup'],
                             treatment_cost=np.array([1, 2]))

    def test_invalid_cost_list_length(self, cost_manager):
        expected = ("treatment_cost must be a list with the same number of "
                    "elements as treatment_features. "
                    "Length of treatment_cost was 2, expected 3.")
        with pytest.raises(UserConfigValidationException, match=expected):
            cost_manager.add(['AveRooms', 'Population', 'AveOccup'],
                             treatment_cost=[1, 2])

    def test_constant_cost_per_treatment_feature(self, cost_manager):
        with patch.object(cost_manager, '_create_policy', return_value=None)\
                as mock_create:
            try:
                cost_manager.add(['AveRooms', 'Population', 'AveOccup'],
                                 treatment_cost=[1, 2, 3])
                cost_manager.compute()
            except TypeError:
                pass
            mock_create.assert_any_call(ANY, ANY, 'AveRooms', 1, ANY, ANY, ANY)
            mock_create.assert_any_call(ANY, ANY, 'Population', 2,
                                        ANY, ANY, ANY)
            mock_create.assert_any_call(ANY, ANY, 'AveOccup', 3, ANY, ANY, ANY)

    def test_per_sample_costs(self, cost_manager):
        with patch.object(cost_manager, '_create_policy', return_value=None)\
                as mock_create:
            costs = [
                [1, 2, 3, 4, 5, 6, 7],
                [2, 3, 4, 5, 6, 7, 1],
            ]
            try:
                cost_manager.add(['AveRooms', 'Population'],
                                 treatment_cost=costs)
                cost_manager.compute()
            except TypeError:
                pass
            mock_create.assert_any_call(
                ANY, ANY, 'AveRooms', [1, 2, 3, 4, 5, 6, 7], ANY, ANY, ANY)
            mock_create.assert_any_call(
                ANY, ANY, 'Population', [2, 3, 4, 5, 6, 7, 1], ANY, ANY, ANY)

    def test_invalid_per_sample_length(self, cost_manager):
        costs = [
            np.array([1, 2, 3, 4, 5]),
            np.array([2, 3, 4, 5, 1]),
        ]
        with pytest.raises(Exception):
            cost_manager.add(['AveRooms', 'Population'], treatment_cost=costs,
                             skip_cat_limit_checks=True)
            cost_manager.compute()


class TestCausalDashboardData:
    def test_categorical_policy(self, housing_data_categorical):
        train_df, test_df, target_feature = housing_data_categorical
        categoricals = train_df.select_dtypes(include=[object]).columns

        # Just use categoricals to force categorical policy tree
        new_features = list(categoricals) + [target_feature]
        train_df = train_df[new_features]
        test_df = test_df[new_features]

        # Sample data for easier debug
        test_df = test_df[:20]

        manager = CausalManager(train_df, test_df, target_feature,
                                ModelTask.REGRESSION, categoricals)

        result = manager.add(['HouseAge_CAT', 'Population_CAT'],
                             random_state=42)
        manager.compute()
        result = manager.get()[0]
        dashboard_data = result._get_dashboard_data()

        policies = dashboard_data['policies']
        assert len(policies) > 0
        for policy in policies:
            tree = policy['policy_tree']
            assert not tree['leaf']
            assert tree['feature'] in categoricals
            assert tree['right_comparison'] == 'eq'
            is_very_old_comparison = tree['comparison_value'] == 'very-old'
            is_high_comparison = tree['comparison_value'] == 'high'
            assert is_very_old_comparison or is_high_comparison



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Responsible AI SDK package."""

# ModelTask is only imported for backwards compatibility
from raiutils.models import ModelTask
from responsibleai.modelanalysis import ModelAnalysis
from responsibleai.rai_insights import RAIInsights

from .__version__ import version
from .feature_metadata import FeatureMetadata

__version__ = version

__all__ = ['ModelAnalysis', 'ModelTask', 'RAIInsights', 'FeatureMetadata']



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.


class UserErrorException(Exception):
    """Exception raised when the user has made an error like
    configuration error.

    :param exception_message: A message describing the error.
    :type exception_message: str
    """
    _error_code = "User Error"


class SystemErrorException(Exception):
    """Exception raised when there is an system error condition
    which is outside the control of the user.

    :param exception_message: A message describing the error.
    :type exception_message: str
    """
    _error_code = "System Error"


class DuplicateManagerConfigException(UserErrorException):
    """An exception indicating that a duplicate configuration
    was detected in any of the RAI managers.

    :param exception_message: A message describing the error.
    :type exception_message: str
    """
    _error_code = 'Duplicate RAI configuration detected'


class UserConfigValidationException(UserErrorException):
    """An exception indicating that some user configuration is not valid.

    :param exception_message: A message describing the error.
    :type exception_message: str
    """
    _error_code = 'Invalid config'


class ConfigAndResultMismatchException(SystemErrorException):
    """An exception indicating that number of configuration and
       results are different.

    :param exception_message: A message describing the error.
    :type exception_message: str
    """
    _error_code = 'Config and result mismatch'


class SchemaErrorException(SystemErrorException):
    """An exception indicating that the schema of the RAI
    component has some error or is not supported.

    :param exception_message: A message describing the error.
    :type exception_message: str
    """
    _error_code = 'Schema Error'



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from responsibleai._interfaces import TaskType
from responsibleai.databalanceanalysis.data_balance_utils import (
    AGGREGATE_BALANCE_MEASURES_KEY, DISTRIBUTION_BALANCE_MEASURES_KEY,
    FEATURE_BALANCE_MEASURES_KEY)
from responsibleai.managers.data_balance_manager import DataBalanceManager
from responsibleai.rai_insights.rai_insights import RAIInsights


class TestDataBalanceManager:
    def test_init_with_valid_input(self, adult_data):
        train_df, test_df, _, target_col, classes = adult_data
        combined = pd.concat([train_df, test_df])

        # Try a combination of train only, test only, and both
        # Since fixtures are being used, easier not to use pytest.parametrize
        combinations = [(train_df, None), (None, test_df), (train_df, test_df)]
        for train, test in combinations:
            combined = pd.concat([train, test])
            manager = DataBalanceManager(
                train=train,
                test=test,
                target_column=target_col,
                classes=classes,
                task_type=TaskType.CLASSIFICATION,
            )
            assert manager._target_column == target_col
            assert manager._train is train
            assert manager._test is test
            assert_frame_equal(
                manager._df.reset_index(drop=True),
                combined.reset_index(drop=True),
            )

    def test_init_with_invalid_input(self):
        manager = DataBalanceManager(
            train=None,
            test=None,
            target_column=None,
            classes=None,
            task_type=None,
        )
        assert manager._target_column is None
        assert manager._classes is None
        assert manager._task_type is None
        assert manager._train is None
        assert manager._test is None
        assert manager._df is None

    @pytest.mark.parametrize("target_col", [None, ""])
    @pytest.mark.parametrize("cols_of_interest", [None, []])
    def test_add_errors_on_invalid_input_basic(
        self, adult_data, target_col, cols_of_interest
    ):
        train_df, test_df, _, _, classes = adult_data

        with pytest.raises(ValueError):
            manager = DataBalanceManager(
                train=train_df,
                test=test_df,
                target_column=target_col,
                classes=classes,
                task_type=TaskType.CLASSIFICATION,
            )
            manager.add(cols_of_interest=cols_of_interest)

    def test_add_errors_on_invalid_input_advanced(self, adult_data):
        train_df, test_df, categorical_cols, target_col, classes = adult_data

        # train and test not specified
        with pytest.raises(ValueError):
            manager = DataBalanceManager(
                train=None,
                test=None,
                target_column=target_col,
                classes=classes,
                task_type=TaskType.CLASSIFICATION,
            )
            manager.add(cols_of_interest=categorical_cols)

        # task_type is not classification
        with pytest.raises(ValueError):
            manager = DataBalanceManager(
                train=train_df,
                test=test_df,
                target_column=target_col,
                classes=classes,
                task_type=TaskType.REGRESSION,
            )
            manager.add(cols_of_interest=categorical_cols)

    def test_validate_with_valid_input(self, adult_data):
        train_df, test_df, categorical_cols, target_col, classes = adult_data
        manager = DataBalanceManager(
            train=train_df,
            test=test_df,
            target_column=target_col,
            classes=classes,
            task_type=TaskType.CLASSIFICATION,
        )
        manager._cols_of_interest = categorical_cols
        manager._validate()  # should not raise any exceptions

    @pytest.mark.parametrize("target_col", [None, ""])
    def test_validate_with_invalid_target_col(self, adult_data, target_col):
        train_df, test_df, _, _, classes = adult_data

        with pytest.raises(ValueError):
            manager = DataBalanceManager(
                train=train_df,
                test=test_df,
                target_column=target_col,
                classes=classes,
                task_type=TaskType.CLASSIFICATION,
            )
            manager._validate()

    @pytest.mark.parametrize("classes", [None, []])
    def test_validate_with_invalid_classes(self, adult_data, classes):
        train_df, test_df, _, target_col, _ = adult_data

        with pytest.raises(ValueError):
            manager = DataBalanceManager(
                train=train_df,
                test=test_df,
                target_column=target_col,
                classes=classes,
                task_type=TaskType.CLASSIFICATION,
            )
            manager._validate()

    def test_validate_with_invalid_input_advanced(self, adult_data):
        train, test, _, target_col, classes = adult_data
        # train and test not specified
        with pytest.raises(ValueError):
            manager = DataBalanceManager(
                train=None,
                test=None,
                target_column=target_col,
                classes=classes,
                task_type=TaskType.CLASSIFICATION,
            )
            manager._validate()

        # task_type is not classification
        with pytest.raises(ValueError):
            manager = DataBalanceManager(
                train=train,
                test=test,
                target_column=target_col,
                classes=classes,
                task_type=TaskType.REGRESSION,
            )
            manager._validate()

    def test_compute_transforms_and_sets_data_balance_measures(
        self, adult_data
    ):
        train_df, test_df, categorical_cols, target_col, classes = adult_data

        manager = DataBalanceManager(
            train=train_df,
            test=test_df,
            target_column=target_col,
            classes=classes,
            task_type=TaskType.CLASSIFICATION,
        )
        manager.add(cols_of_interest=categorical_cols)

        assert manager._data_balance_measures is None
        manager.compute()
        assert manager._data_balance_measures is not None

    def test_compute_with_col_less_than_two_unique_values(
        self, singular_valued_col_data
    ):
        # col1 has only one unique value, so we should get a warning per class
        # that feature balance measures weren't computed for this col
        cols_of_interest = ["col1", "col2"]
        target = "target"
        classes = [0, 1]

        manager = DataBalanceManager(
            train=singular_valued_col_data[:3],
            test=singular_valued_col_data[3:],
            target_column=target,
            classes=classes,
            task_type=TaskType.CLASSIFICATION,
        )
        manager.add(cols_of_interest=cols_of_interest)

        with pytest.warns(UserWarning) as warninfo:
            manager.compute()

        warns = {(warn.category, warn.message.args[0]) for warn in warninfo}

        warn_text = (f"Column '{cols_of_interest[0]}' has less than 2 unique"
                     f" values, so feature balance measures for {target}=="
                     "{pos_label} are not being computed for this column.")
        expected = {
            (UserWarning, warn_text.format(pos_label=classes[0])),
            (UserWarning, warn_text.format(pos_label=classes[1])),
        }
        for expected_warning in expected:
            assert expected_warning in warns

    def test_sets_data_balance_measures(
        self,
        adult_data_feature_balance_measures,
        adult_data_distribution_balance_measures,
        adult_data_aggregate_balance_measures,
    ):
        manager = DataBalanceManager(
            target_column=None,
            classes=None,
            train=None,
            test=None,
            task_type=None,
        )
        assert manager._data_balance_measures is None

        manager._set_data_balance_measures(
            adult_data_feature_balance_measures,
            adult_data_distribution_balance_measures,
            adult_data_aggregate_balance_measures,
        )
        assert manager._data_balance_measures is not None

        d = manager._data_balance_measures
        assert isinstance(d, dict)
        assert FEATURE_BALANCE_MEASURES_KEY in d
        assert DISTRIBUTION_BALANCE_MEASURES_KEY in d
        assert AGGREGATE_BALANCE_MEASURES_KEY in d

    def test_save_and_load_basic(self, tmpdir, adult_data):
        train_df, test_df, categorical_cols, target_col, classes = adult_data
        task_type = TaskType.CLASSIFICATION

        saved = DataBalanceManager(
            train=train_df,
            test=test_df,
            target_column=target_col,
            classes=classes,
            task_type=task_type,
        )
        assert saved._is_added is False

        save_dir = tmpdir.mkdir("save-dir")
        saved._save(save_dir)  # should not create any configs

        rai_insights = RAIInsights(
            model=None,
            train=train_df,
            test=test_df,
            target_column=target_col,
            task_type="classification",
            categorical_features=categorical_cols,
        )
        loaded = saved._load(save_dir, rai_insights)

        assert saved._target_column == loaded._target_column == target_col
        assert saved._classes == loaded._classes == list(map(str, classes))
        assert saved._task_type == loaded._task_type == task_type
        assert_frame_equal(saved._train, loaded._train)
        assert_frame_equal(saved._test, loaded._test)
        assert saved._is_added is loaded._is_added is False

        # All the instances variables from RAIInsights get set.
        assert_frame_equal(loaded._train, saved._train)
        assert_frame_equal(loaded._test, saved._test)
        assert loaded._target_column == saved._target_column
        # Also, df gets created using rai_insights.train and rai_insights.test
        assert_frame_equal(loaded._df, saved._df)

        # All the instance variables specific to the manager, such as
        # _cols_of_interest, don't get set.
        assert loaded._is_added is False
        assert loaded._cols_of_interest is None
        assert loaded._data_balance_measures is None

    def test_save_and_load_with_add(self, tmpdir, adult_data):
        train_df, test_df, categorical_cols, target_col, classes = adult_data
        task_type = TaskType.CLASSIFICATION

        saved = DataBalanceManager(
            train=train_df,
            test=test_df,
            target_column=target_col,
            classes=classes,
            task_type=task_type,
        )
        saved.add(cols_of_interest=categorical_cols)

        save_dir = tmpdir.mkdir("save-dir")
        saved._save(save_dir)

        rai_insights = RAIInsights(
            model=None,
            train=train_df,
            test=test_df,
            target_column=target_col,
            task_type="classification",
            categorical_features=categorical_cols,
        )
        loaded = saved._load(save_dir, rai_insights)

        assert saved._target_column == loaded._target_column == target_col
        assert saved._classes == loaded._classes == list(map(str, classes))
        assert saved._task_type == loaded._task_type == task_type
        assert_frame_equal(saved._train, loaded._train)
        assert_frame_equal(saved._test, loaded._test)
        assert saved._is_added is loaded._is_added is True
        assert_frame_equal(saved._df, loaded._df)

        assert saved._cols_of_interest == loaded._cols_of_interest

        # Populated in compute() so should be None
        assert (
            saved._data_balance_measures
            is loaded._data_balance_measures
            is None
        )

    def test_save_and_load_with_add_and_compute(self, tmpdir, adult_data):
        train_df, test_df, categorical_cols, target_col, classes = adult_data
        task_type = TaskType.CLASSIFICATION

        saved = DataBalanceManager(
            train=train_df,
            test=test_df,
            target_column=target_col,
            classes=classes,
            task_type=task_type,
        )
        saved.add(cols_of_interest=categorical_cols)
        saved.compute()

        save_dir = tmpdir.mkdir("save-dir")
        saved._save(save_dir)

        rai_insights = RAIInsights(
            model=None,
            train=train_df,
            test=test_df,
            target_column=target_col,
            task_type="classification",
            categorical_features=categorical_cols,
        )
        loaded = saved._load(save_dir, rai_insights)

        assert saved._target_column == loaded._target_column == target_col
        assert saved._classes == loaded._classes == list(map(str, classes))
        assert saved._task_type == loaded._task_type == task_type
        assert_frame_equal(saved._train, loaded._train)
        assert_frame_equal(saved._test, loaded._test)
        assert saved._is_added is loaded._is_added is True
        assert_frame_equal(saved._df, loaded._df)

        assert saved._cols_of_interest == loaded._cols_of_interest

        assert saved._data_balance_measures == loaded._data_balance_measures



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Implementation of Model Analysis API."""

# ModelTask is only imported for backwards compatibility
from raiutils.models import ModelTask
from responsibleai.modelanalysis.model_analysis import ModelAnalysis

__all__ = ["ModelAnalysis", "ModelTask"]



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import timeit
from typing import List

import numpy as np
import pandas as pd


def _measure_time(manager_compute_func):
    def compute_wrapper(*args, **kwargs):
        _separator(80)
        start_time = timeit.default_timer()
        manager_compute_func(*args, **kwargs)
        elapsed = timeit.default_timer() - start_time
        m, s = divmod(elapsed, 60)
        print('Time taken: {0} min {1} sec'.format(
              m, s))
        _separator(80)
    return compute_wrapper


def _separator(max_len):
    print('=' * max_len)


def _find_features_having_missing_values(
        data: pd.DataFrame) -> List[str]:
    """Return list of features which have missing values.

    :param data: The dataset to check.
    :type data: pd.Dataframe
    :return: List of feature names which have missing values.
    :rtype: List[str]
    """
    list_of_feature_having_missing_values = []
    for feature in data.columns.tolist():
        if np.any(data[feature].isnull()):
            list_of_feature_having_missing_values.append(feature)
    return list_of_feature_having_missing_values



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import pytest

from raiutils.exceptions import UserConfigValidationException
from raiutils.models import ModelTask
from responsibleai._internal.constants import ListProperties, ManagerNames

LIGHTGBM_METHOD = 'mimic.lightgbm'


def setup_explainer(rai_insights, add_explainer=True):
    if add_explainer:
        if rai_insights.model is None:
            with pytest.raises(
                    UserConfigValidationException,
                    match='Model is required for model explanations'):
                rai_insights.explainer.add()
            return
        else:
            rai_insights.explainer.add()
        # Validate calling add multiple times prints a warning
        with pytest.warns(
            UserWarning,
            match="DUPLICATE-EXPLAINER-CONFIG: Ignoring. "
                  "Explanation has already been added, "
                  "currently limited to one explainer type."):
            rai_insights.explainer.add()
    rai_insights.explainer.compute()


def validate_explainer(rai_insights, X_train, X_test, classes):
    if rai_insights.model is None:
        return
    explanations = rai_insights.explainer.get()
    assert isinstance(explanations, list)
    assert len(explanations) == 1
    explanation = explanations[0]
    if rai_insights._feature_metadata is not None and \
            rai_insights._feature_metadata.dropped_features is not None:
        num_cols = len(X_train.columns) - 1 - len(
            rai_insights._feature_metadata.dropped_features)
    else:
        num_cols = len(X_train.columns) - 1
    if classes is not None:
        assert len(explanation.local_importance_values) == len(classes)
        assert len(explanation.local_importance_values[0]) == len(X_test)
        assert len(explanation.local_importance_values[0][0]) == num_cols
    else:
        assert len(explanation.local_importance_values) == len(X_test)
        assert len(explanation.local_importance_values[0]) == num_cols

    properties = rai_insights.explainer.list()
    assert properties[ListProperties.MANAGER_TYPE] == ManagerNames.EXPLAINER
    assert 'id' in properties
    assert properties['method'] == LIGHTGBM_METHOD
    if classes is not None:
        assert properties['model_task'] == ModelTask.CLASSIFICATION
    else:
        assert properties['model_task'] == ModelTask.REGRESSION
    assert properties['model_type'] is None
    assert properties['is_raw'] is False
    assert properties['is_engineered'] is False

    # Check the internal state of explainer manager
    assert rai_insights.explainer._is_added
    assert rai_insights.explainer._is_run



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import json
import random
import ssl

import numpy as np
import pandas as pd
import pytest
from dice_ml.utils import helpers
# Defines common utilities for responsibleai tests
from sklearn.model_selection import train_test_split

from rai_test_utils.datasets.tabular import \
    create_iris_data as _create_iris_data
from raiutils.common.retries import retry_function

ADULT_DROPPED_FEATURES = ['workclass', 'education', 'marital_status',
                          'occupation', 'race', 'gender']
ADULT_CATEGORICAL_FEATURES_AFTER_DROP = ['age', 'hours_per_week']


def create_iris_data():
    (
        X_train,
        X_test,
        y_train,
        y_test,
        feature_names,
        classes,
    ) = _create_iris_data()

    # create duplicate index
    test_index_list = X_test.index.tolist()
    test_index_list.append(test_index_list[0])
    X_test = X_test.reindex(test_index_list)
    y_test_list = y_test.tolist()
    y_test_list.append(y_test[0])
    y_test = np.array(y_test_list)

    return X_train, X_test, y_train, y_test, feature_names, classes


def create_tiny_forecasting_dataset():
    X_train = pd.DataFrame({
        "time": [
            pd.Timestamp("2023-08-18T06:00:00"),
            pd.Timestamp("2023-08-18T07:00:00"),
            pd.Timestamp("2023-08-18T08:00:00"),
            pd.Timestamp("2023-08-18T09:00:00")
        ],
        "id": [1, 1, 1, 1]
    })
    X_test = pd.DataFrame({
        "time": [
            pd.Timestamp("2023-08-18T10:00:00"),
            pd.Timestamp("2023-08-18T11:00:00"),
            pd.Timestamp("2023-08-18T12:00:00"),
            pd.Timestamp("2023-08-18T13:00:00")
        ],
        "id": [1, 1, 1, 1]
    })
    y_train = [0, 1, 2, 3]
    y_test = [4, 5, 6, 7]
    return X_train, X_test, y_train, y_test


class RandomForecastingModel():
    def forecast(self, X):
        return np.array([random.random() for _ in range(len(X))])


class RandomForecastingModelWithQuantiles(RandomForecastingModel):
    def forecast_quantiles(self, X, quantiles):
        return [
            [random.random() for _ in range(len(X))],
            [random.random() for _ in range(len(X))]
        ]


class FetchDiceAdultCensusIncomeDataset(object):
    def __init__(self):
        pass

    def fetch(self):
        return helpers.load_adult_income_dataset()


def create_adult_income_dataset(create_small_dataset=True):
    # workaround for SSL expiration error from UCI website
    default_http_context = ssl._create_default_https_context
    ssl._create_default_https_context = ssl._create_unverified_context
    fetcher = FetchDiceAdultCensusIncomeDataset()
    action_name = "Adult dataset download"
    err_msg = "Failed to download adult dataset"
    max_retries = 4
    retry_delay = 60
    dataset = retry_function(
        fetcher.fetch,
        action_name,
        err_msg,
        max_retries=max_retries,
        retry_delay=retry_delay,
    )
    continuous_features = ["age", "hours_per_week"]
    target_name = "income"
    target = dataset[target_name]
    classes = list(np.unique(target))
    feature_columns = dataset.drop(columns=[target_name]).columns.tolist()
    feature_range_keys = ["column_name", "range_type", "unique_values"]
    categorical_features = list(
        set(dataset.columns) - set(continuous_features) - set([target_name])
    )
    # Split data into train and test
    if create_small_dataset:
        data_train, data_test, y_train, y_test = train_test_split(
            dataset,
            target,
            train_size=500,
            test_size=50,
            random_state=7,
            stratify=target,
        )
    else:
        data_train, data_test, y_train, y_test = train_test_split(
            dataset, target, test_size=5000, random_state=7, stratify=target
        )
    ssl._create_default_https_context = default_http_context
    return (
        data_train,
        data_test,
        y_train,
        y_test,
        categorical_features,
        continuous_features,
        target_name,
        classes,
        feature_columns,
        feature_range_keys,
    )


def assert_series_and_dict_equal(left: pd.Series, right: dict):
    left_json = left.to_json(orient="index")
    left_dict = json.loads(left_json)
    assert left_dict == pytest.approx(right)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines the Counterfactual Manager class."""
import json
import uuid
import warnings
from pathlib import Path
from typing import Any, List, Optional, Union

import dice_ml
import jsonschema
import numpy as np
import pandas as pd
from dice_ml import Dice
from dice_ml.counterfactual_explanations import CounterfactualExplanations
from dice_ml.explainer_interfaces.explainer_base import ExplainerBase

from raiutils.exceptions import UserConfigValidationException
from raiutils.models import ModelTask
from responsibleai._config.base_config import BaseConfig
from responsibleai._data_validations import validate_train_test_categories
from responsibleai._interfaces import CounterfactualData
from responsibleai._internal.constants import (CounterfactualManagerKeys,
                                               FileFormats, ListProperties,
                                               ManagerNames)
from responsibleai._tools.shared.state_directory_management import \
    DirectoryManager
from responsibleai.exceptions import (DuplicateManagerConfigException,
                                      SchemaErrorException)
from responsibleai.managers.base_manager import BaseManager
from responsibleai.utils import (_find_features_having_missing_values,
                                 _measure_time)


class CounterfactualConstants:
    OPPOSITE = 'opposite'
    CLASSIFIER = 'classifier'
    REGRESSOR = 'regressor'
    SKLEARN = 'sklearn'
    RANDOM = 'random'


class _CommonSchemaConstants:
    LOCAL_IMPORTANCE = 'local_importance'
    METADATA = 'metadata'
    SUMMARY_IMPORTANCE = 'summary_importance'
    VERSION = 'version'


class _V1SchemaConstants:
    CF_EXAMPLES_LIST = 'cf_examples_list'
    LOCAL_IMPORTANCE = _CommonSchemaConstants.LOCAL_IMPORTANCE
    METADATA = _CommonSchemaConstants.METADATA
    SUMMARY_IMPORTANCE = _CommonSchemaConstants.SUMMARY_IMPORTANCE

    ALL = [CF_EXAMPLES_LIST, LOCAL_IMPORTANCE, METADATA, SUMMARY_IMPORTANCE]


class _V2SchemaConstants:
    CFS_LIST = 'cfs_list'
    DATA_INTERFACE = 'data_interface'
    DESIRED_CLASS = 'desired_class'
    DESIRED_RANGE = 'desired_range'
    FEATURE_NAMES = 'feature_names'
    FEATURE_NAMES_INCLUDING_TARGET = 'feature_names_including_target'
    LOCAL_IMPORTANCE = _CommonSchemaConstants.LOCAL_IMPORTANCE
    METADATA = _CommonSchemaConstants.METADATA
    MODEL_TYPE = 'model_type'
    SUMMARY_IMPORTANCE = _CommonSchemaConstants.SUMMARY_IMPORTANCE
    TEST_DATA = 'test_data'

    ALL = [CFS_LIST, DATA_INTERFACE, DESIRED_CLASS, DESIRED_RANGE,
           FEATURE_NAMES, FEATURE_NAMES_INCLUDING_TARGET,
           LOCAL_IMPORTANCE, METADATA, MODEL_TYPE,
           SUMMARY_IMPORTANCE, TEST_DATA]


class _SchemaVersions:
    V1 = '1.0'
    V2 = '2.0'

    ALL_VERSIONS = [V1, V2]


def _get_schema_version(counterfactuals_dict):
    """
    Get the version from the serialized version of the
    counterfactual examples.

    :param counterfactuals_dict: Serialized version of the
                                 counterfactual example.
    :type counterfactuals_dict: Dict
    """
    if _CommonSchemaConstants.METADATA not in counterfactuals_dict:
        raise SchemaErrorException(
            "No metadata field found in serialized counterfactual output"
        )

    metadata = counterfactuals_dict[_CommonSchemaConstants.METADATA]
    if _CommonSchemaConstants.VERSION not in metadata:
        raise SchemaErrorException(
            "No version field found in serialized counterfactual output"
        )

    version = metadata[_CommonSchemaConstants.VERSION]
    if version not in _SchemaVersions.ALL_VERSIONS:
        raise SchemaErrorException(
            "Schema version {0} not supported."
            "Support versions are {1}".format(
                version, ','.join(_SchemaVersions.ALL_VERSIONS)
            )
        )

    return version


class CounterfactualConfig(BaseConfig):
    """Defines the configuration for generating counterfactuals."""
    METHOD = 'method'
    CONTINUOUS_FEATURES = 'continuous_features'
    TOTAL_CFS = 'total_CFs'
    ID = 'id'
    DESIRED_RANGE = 'desired_range'
    DESIRED_CLASS = 'desired_class'
    PERMITTED_RANGE = 'permitted_range'
    FEATURES_TO_VARY = 'features_to_vary'
    FEATURE_IMPORTANCE = 'feature_importance'

    IS_COMPUTED = 'is_computed'
    COUNTERFACTUAL_OBJ = 'counterfactual_obj'
    HAS_COMPUTATION_FAILED = 'has_computation_failed'
    FAILURE_REASON = 'failure_reason'

    CONFIG_FILE_NAME = f'config{FileFormats.JSON}'
    RESULT_FILE_NAME = f'result{FileFormats.JSON}'
    EXPLAINER_FILE_NAME = f'explainer{FileFormats.PKL}'

    def __init__(self, method, continuous_features, total_CFs,
                 desired_class=CounterfactualConstants.OPPOSITE,
                 desired_range=None, permitted_range=None,
                 features_to_vary=None, feature_importance=False,
                 id=None):
        super(CounterfactualConfig, self).__init__()
        self.method = method
        self.continuous_features = continuous_features
        self.total_CFs = total_CFs
        if id is None:
            self.id = str(uuid.uuid4())
        else:
            self.id = id
        self.desired_range = desired_range
        self.desired_class = desired_class
        self.permitted_range = permitted_range
        self.features_to_vary = features_to_vary
        self.feature_importance = feature_importance
        self.is_computed = False
        self.counterfactual_obj = None
        self.has_computation_failed = False
        self.failure_reason = None
        self.explainer = None

    def __eq__(self, other_cf_config):
        return (
            self.method == other_cf_config.method and
            self.continuous_features == other_cf_config.continuous_features and
            self.total_CFs == other_cf_config.total_CFs and
            self.desired_range == other_cf_config.desired_range and
            self.desired_class == other_cf_config.desired_class and
            self.permitted_range == other_cf_config.permitted_range and
            self.features_to_vary == other_cf_config.features_to_vary and
            self.feature_importance == other_cf_config.feature_importance
        )

    def get_config_as_dict(self):
        """Returns the dictionary representation of configuration
           in the CounterfactualConfig.

        The dictionary contains the counterfactual method, continuous features,
        total counterfactuals, desired class, desired range,
        permitted range, features to vary and feature importance.

        :return: The dictionary representation of the CounterfactualConfig.
        :rtype: dict
        """
        return {
            CounterfactualConfig.METHOD: self.method,
            CounterfactualConfig.CONTINUOUS_FEATURES: self.continuous_features,
            CounterfactualConfig.TOTAL_CFS: self.total_CFs,
            CounterfactualConfig.ID: self.id,
            CounterfactualConfig.DESIRED_RANGE: self.desired_range,
            CounterfactualConfig.DESIRED_CLASS: self.desired_class,
            CounterfactualConfig.PERMITTED_RANGE: self.permitted_range,
            CounterfactualConfig.FEATURES_TO_VARY: self.features_to_vary,
            CounterfactualConfig.FEATURE_IMPORTANCE: self.feature_importance
        }

    def save_config(self, config_directory_path):
        config_file_path = (config_directory_path /
                            CounterfactualConfig.CONFIG_FILE_NAME)
        with open(config_file_path, 'w') as config_file:
            json.dump(
                self.get_config_as_dict(), config_file)

    @staticmethod
    def load_config(config_directory_path):
        config_path = (config_directory_path /
                       CounterfactualConfig.CONFIG_FILE_NAME)
        with open(config_path, 'r') as config_file:
            cf_config = json.load(config_file)

        counterfactual_config = CounterfactualConfig(
            method=cf_config[CounterfactualConfig.METHOD],
            continuous_features=cf_config[
                CounterfactualConfig.CONTINUOUS_FEATURES],
            total_CFs=cf_config[CounterfactualConfig.TOTAL_CFS],
            desired_class=cf_config[CounterfactualConfig.DESIRED_CLASS],
            desired_range=cf_config[CounterfactualConfig.DESIRED_RANGE],
            permitted_range=cf_config[
                CounterfactualConfig.PERMITTED_RANGE],
            features_to_vary=cf_config[
                CounterfactualConfig.FEATURES_TO_VARY],
            feature_importance=cf_config[
                CounterfactualConfig.FEATURE_IMPORTANCE],
            id=cf_config[CounterfactualConfig.ID])

        return counterfactual_config

    def get_result(self, include_counterfactual_obj=True):
        """Returns the dictionary representation of result of the computation
           in the CounterfactualConfig.

        :param include_counterfactual_obj: Whether to include serialized
                                           version of counterfactual object.
        :type method: bool
        :return: The dictionary representation of result in the
                 CounterfactualConfig.
        :rtype: dict
        """
        result = {}
        if include_counterfactual_obj:
            if self.counterfactual_obj is not None:
                result[CounterfactualConfig.COUNTERFACTUAL_OBJ] = \
                    self.counterfactual_obj.to_json()
            else:
                result[CounterfactualConfig.COUNTERFACTUAL_OBJ] = None

        result[CounterfactualConfig.HAS_COMPUTATION_FAILED] = \
            self.has_computation_failed
        result[CounterfactualConfig.FAILURE_REASON] = self.failure_reason
        result[CounterfactualConfig.IS_COMPUTED] = self.is_computed

        return result

    def save_result(self, data_directory_path):
        cf_result = self.get_result()
        if cf_result[CounterfactualConfig.COUNTERFACTUAL_OBJ] is not None:
            counterfactuals_dict = json.loads(
                cf_result[CounterfactualConfig.COUNTERFACTUAL_OBJ]
            )

            schema_version = _get_schema_version(counterfactuals_dict)
            if schema_version == _SchemaVersions.V1:
                cf_schema_keys = _V1SchemaConstants.ALL
            else:
                cf_schema_keys = _V2SchemaConstants.ALL

            for counterfactual_examples_key in cf_schema_keys:
                file_path = (data_directory_path /
                             (counterfactual_examples_key + FileFormats.JSON))
                with open(file_path, 'w') as file_path:
                    json.dump(
                        counterfactuals_dict[counterfactual_examples_key],
                        file_path)

        file_path = data_directory_path / (
            CounterfactualConfig.HAS_COMPUTATION_FAILED + FileFormats.JSON)
        with open(file_path, 'w') as file_path:
            json.dump(
                cf_result[CounterfactualConfig.HAS_COMPUTATION_FAILED],
                file_path)

        file_path = (data_directory_path /
                     (CounterfactualConfig.FAILURE_REASON + FileFormats.JSON))
        with open(file_path, 'w') as file_path:
            json.dump(
                cf_result[CounterfactualConfig.FAILURE_REASON],
                file_path)

        file_path = (data_directory_path /
                     (CounterfactualConfig.IS_COMPUTED + FileFormats.JSON))
        with open(file_path, 'w') as file_path:
            json.dump(
                cf_result[CounterfactualConfig.IS_COMPUTED],
                file_path)

    def load_result(self, data_directory_path):
        metadata_file_path = (
            data_directory_path /
            (_CommonSchemaConstants.METADATA + FileFormats.JSON))

        if metadata_file_path.exists():
            with open(metadata_file_path, 'r') as result_file:
                metadata = json.load(result_file)

            if metadata['version'] == _SchemaVersions.V1:
                cf_schema_keys = _V1SchemaConstants.ALL
            else:
                cf_schema_keys = _V2SchemaConstants.ALL

            counterfactual_examples_dict = {}
            for counterfactual_examples_key in cf_schema_keys:
                result_path = (
                    data_directory_path /
                    (counterfactual_examples_key + FileFormats.JSON))
                with open(result_path, 'r') as result_file:
                    counterfactual_examples_dict[
                        counterfactual_examples_key] = json.load(result_file)

            counterfactuals_json_str = json.dumps(counterfactual_examples_dict)
            self.counterfactual_obj = \
                CounterfactualExplanations.from_json(counterfactuals_json_str)
        else:
            self.counterfactual_obj = None

        result_path = (
            data_directory_path /
            (CounterfactualConfig.HAS_COMPUTATION_FAILED + FileFormats.JSON))
        with open(result_path, 'r') as result_file:
            self.has_computation_failed = json.load(result_file)

        result_path = (
            data_directory_path /
            (CounterfactualConfig.FAILURE_REASON + FileFormats.JSON))
        with open(result_path, 'r') as result_file:
            self.failure_reason = json.load(result_file)

        result_path = (
            data_directory_path /
            (CounterfactualConfig.IS_COMPUTED + FileFormats.JSON))
        with open(result_path, 'r') as result_file:
            self.is_computed = json.load(result_file)

    def save_explainer(self, explainer_directory_path):
        file_path = (explainer_directory_path /
                     CounterfactualConfig.EXPLAINER_FILE_NAME)
        try:
            self.explainer.serialize_explainer(file_path)
        except Exception:
            pass

    def load_explainer(self, explainer_directory_path):
        file_path = (explainer_directory_path /
                     CounterfactualConfig.EXPLAINER_FILE_NAME)
        try:
            self.explainer = ExplainerBase.deserialize_explainer(file_path)
        except Exception:
            pass


class CounterfactualManager(BaseManager):
    """Defines the CounterfactualManager for generating counterfactuals
        from a model.
    """
    _TRAIN = '_train'
    _TEST = '_test'
    _MODEL = '_model'
    _TARGET_COLUMN = '_target_column'
    _TASK_TYPE = '_task_type'
    _CATEGORICAL_FEATURES = '_categorical_features'
    _COUNTERFACTUAL_CONFIG_LIST = '_counterfactual_config_list'

    def __init__(self, model: Any, train: pd.DataFrame, test: pd.DataFrame,
                 target_column: str, task_type: str,
                 categorical_features: List[str]):
        """Creates a CounterfactualManager object.

        :param model: The model to generate counterfactuals from.
            A model that implements sklearn.predict or sklearn.predict_proba
            or function that accepts a 2d ndarray.
        :type model: object
        :param initialization_examples: A matrix of feature vector
            examples (# examples x # features) for initializing the explainer.
        :type initialization_examples: pandas.DataFrame
        :param evaluation_examples: A matrix of feature vector
            examples (# examples x # features) on which to explain the
            model's output.
        :type evaluation_examples: pandas.DataFrame
        :param target_column: The name of the label column.
        :type target_column: str
        :param task_type: Task type is either 'classification/regression'
        :type task_type: str
        :param categorical_features: The categorical feature names.
        :type categorical_features: list[str]
        """
        self._model = model
        self._train = train
        self._test = test
        self._target_column = target_column
        self._task_type = task_type
        self._categorical_features = categorical_features
        self._counterfactual_config_list = []

    def _create_diceml_explainer(self, method, continuous_features):

        dice_data = dice_ml.Data(dataframe=self._train,
                                 continuous_features=continuous_features,
                                 outcome_name=self._target_column)
        model_type = (
            CounterfactualConstants.CLASSIFIER
            if self._task_type == ModelTask.CLASSIFICATION else
            CounterfactualConstants.REGRESSOR)
        dice_model = dice_ml.Model(model=self._model,
                                   backend=CounterfactualConstants.SKLEARN,
                                   model_type=model_type)

        dice_explainer = Dice(dice_data, dice_model, method=method)

        return dice_explainer

    def _add_counterfactual_config(self, new_counterfactual_config):
        if any(_find_features_having_missing_values(self._train)):
            raise UserConfigValidationException(
                'Missing values are not allowed in the '
                'train dataset while computing counterfactuals.')
        if any(_find_features_having_missing_values(self._test)):
            raise UserConfigValidationException(
                'Missing values are not allowed in the '
                'test dataset while computing counterfactuals.')

        if self._model is None:
            raise UserConfigValidationException(
                'Model is required for counterfactual example generation and '
                'feature importances')

        validate_train_test_categories(
            train_data=self._train,
            test_data=self._test,
            rai_compute_type='Counterfactual example generation',
            categoricals=self._categorical_features)

        to_vary = new_counterfactual_config.features_to_vary
        if to_vary != 'all':
            difference_set = set(to_vary) - set(self._train.columns)
            if len(difference_set) > 0:
                message = ("Feature names in features_to_vary do "
                           f"not exist in train data: {list(difference_set)}")
                raise UserConfigValidationException(message)

        if new_counterfactual_config.permitted_range is not None:
            permitted_features = \
                list(new_counterfactual_config.permitted_range)
            difference_set = set(permitted_features) - set(self._train.columns)
            if len(difference_set) > 0:
                message = ("Feature names in permitted_range do "
                           f"not exist in train data: {list(difference_set)}")
                raise UserConfigValidationException(message)

        if self._task_type == ModelTask.CLASSIFICATION:
            if new_counterfactual_config.desired_class is None:
                raise UserConfigValidationException(
                    'The desired_class attribute should be either \'{0}\''
                    ' for binary classification or the class value for '
                    'multi-classification scenarios.'.format(
                        CounterfactualConstants.OPPOSITE))

            is_multiclass = len(np.unique(
                self._train[self._target_column].values).tolist()) > 2
            if is_multiclass and \
                    new_counterfactual_config.desired_class == \
                    CounterfactualConstants.OPPOSITE:
                raise UserConfigValidationException(
                    'The desired_class attribute should not be \'{0}\''
                    ' It should be the class value for multiclass'
                    ' classification scenario.'.format(
                        CounterfactualConstants.OPPOSITE))

        if self._task_type == ModelTask.REGRESSION:
            if new_counterfactual_config.desired_range is None:
                raise UserConfigValidationException(
                    'The desired_range should not be None'
                    ' for regression scenarios.')

        if new_counterfactual_config.feature_importance and\
                new_counterfactual_config.total_CFs < 10:
            raise UserConfigValidationException(
                "A total_CFs value of at least 10 is required to "
                "use counterfactual feature importances. "
                "Either increase total_CFs to at least 10 or "
                "set feature_importance to False.")

        is_duplicate = new_counterfactual_config.is_duplicate(
            self._counterfactual_config_list)

        if is_duplicate:
            raise DuplicateManagerConfigException(
                'Duplicate counterfactual configuration detected.')
        else:
            self._counterfactual_config_list.append(new_counterfactual_config)

    def add(self,
            total_CFs: int,
            method: str = CounterfactualConstants.RANDOM,
            desired_class: Optional[Union[str, int]] = None,
            desired_range: Optional[List] = None,
            permitted_range: Optional[dict] = None,
            features_to_vary: Union[str, List[str]] = 'all',
            feature_importance: bool = True):
        """Add a counterfactual generation configuration to be computed later.

        :param total_CFs: Total number of counterfactuals required.
        :type total_CFs: int
        :param method: Type of dice-ml explainer. Either of "random", "genetic"
            or "kdtree".
        :type method: str
        :param desired_class: Desired counterfactual class. For binary
            classification, this needs to be set as "opposite".
        :type desired_class: string or int
        :param desired_range: For regression problems.
            Contains the outcome range to generate counterfactuals in.
        :type desired_range: list
        :param permitted_range: Dictionary with feature names as keys and
            permitted range in list as values. Defaults to the range inferred
            from training data.
        :type permitted_range: dict
        :param features_to_vary: Either a string "all" or a list of
            feature names to vary.
        :type features_to_vary: list
        :param feature_importance: Flag to compute feature importance using
            dice-ml.
        :type feature_importance: bool
        """
        if self._categorical_features is None:
            continuous_features = \
                list(set(self._train.columns) - set([self._target_column]))
        else:
            continuous_features = list(set(self._train.columns) -
                                       set([self._target_column]) -
                                       set(self._categorical_features))

        counterfactual_config = CounterfactualConfig(
            method=method,
            continuous_features=continuous_features,
            total_CFs=total_CFs,
            desired_class=desired_class,
            desired_range=desired_range,
            permitted_range=permitted_range,
            features_to_vary=features_to_vary,
            feature_importance=feature_importance)

        self._add_counterfactual_config(counterfactual_config)

    @_measure_time
    def compute(self):
        """Computes the counterfactual examples by running the counterfactual
           configuration."""
        print("Counterfactual")
        for cf_config in self._counterfactual_config_list:
            if not cf_config.is_computed:
                cf_config.is_computed = True
                try:
                    print("Current Status: Generating {0} counterfactuals"
                          " for {1} samples".format(
                              cf_config.total_CFs, len(self._test)))

                    cf_config.explainer = self._create_diceml_explainer(
                        method=cf_config.method,
                        continuous_features=cf_config.continuous_features)

                    X_test = self._test.drop([self._target_column], axis=1)

                    if not cf_config.feature_importance:
                        counterfactual_obj = \
                            cf_config.explainer.generate_counterfactuals(
                                X_test, total_CFs=cf_config.total_CFs,
                                desired_class=cf_config.desired_class,
                                desired_range=cf_config.desired_range,
                                features_to_vary=cf_config.features_to_vary,
                                permitted_range=cf_config.permitted_range)
                    else:
                        counterfactual_obj = \
                            cf_config.explainer.global_feature_importance(
                                X_test,
                                total_CFs=cf_config.total_CFs,
                                desired_class=cf_config.desired_class,
                                desired_range=cf_config.desired_range,
                                features_to_vary=cf_config.features_to_vary,
                                permitted_range=cf_config.permitted_range)

                    # Validate the serialized output against schema
                    schema = CounterfactualManager._get_counterfactual_schema(
                        version=counterfactual_obj.metadata['version'])
                    jsonschema.validate(
                        json.loads(counterfactual_obj.to_json()), schema)

                    cf_config.counterfactual_obj = counterfactual_obj

                    print('Current Status: Generated {0} counterfactuals'
                          ' for {1} samples.'.format(
                              cf_config.total_CFs, len(self._test)))
                except Exception as e:
                    cf_config.has_computation_failed = True
                    cf_config.failure_reason = str(e)
                    raise e

    def request_counterfactuals(self, query_id: str, data: Any):
        """Return the counterfactuals for a given point.

        :param query_id: The query id for finding the
                         counterfactual config.
        :type query_id: str
        :param data: The data point for which the counterfactuals
                     need to be generated.
        :type data: Any
        :return: An object of type CounterfactualData with
                 counterfactuals for the given data point.
        :rtype: CounterfactualData
        """
        if not isinstance(data, pd.DataFrame):
            raise UserConfigValidationException(
                'Data is of type {0} but it must be '
                'a pandas DataFrame.'.format(type(data)))

        if data.shape[0] > 1:
            raise UserConfigValidationException(
                'Only one row of data is allowed for '
                'counterfactual generation.')

        query_cf_config = None
        for cf_config in self._counterfactual_config_list:
            if cf_config.id == query_id:
                query_cf_config = cf_config
                break

        if query_cf_config is None:
            raise UserConfigValidationException(
                'No counterfactual config found for id {0}.'.format(
                    query_id))

        if not query_cf_config.feature_importance:
            counterfactual_obj = \
                query_cf_config.explainer.generate_counterfactuals(
                    data, total_CFs=query_cf_config.total_CFs,
                    desired_class=query_cf_config.desired_class,
                    desired_range=query_cf_config.desired_range,
                    features_to_vary=query_cf_config.features_to_vary,
                    permitted_range=query_cf_config.permitted_range)
        else:
            counterfactual_obj = \
                query_cf_config.explainer.local_feature_importance(
                    data,
                    total_CFs=query_cf_config.total_CFs,
                    desired_class=query_cf_config.desired_class,
                    desired_range=query_cf_config.desired_range,
                    features_to_vary=query_cf_config.features_to_vary,
                    permitted_range=query_cf_config.permitted_range)

        # Validate the serialized output against schema
        schema = CounterfactualManager._get_counterfactual_schema(
            version=counterfactual_obj.metadata['version'])
        jsonschema.validate(
            json.loads(counterfactual_obj.to_json()), schema)

        return self._get_counterfactual(query_cf_config, counterfactual_obj)

    def get(self, failed_to_compute=False):
        """Return computed counterfactual examples objects or failure reason.

        :param failed_to_compute: Get the failure reasons if counterfactual
                                  examples failed to compute.
        :type failed_to_compute: Optional[bool]
        """
        if not failed_to_compute:
            counterfactual_obj_list = []
            for counterfactual_config in self._counterfactual_config_list:
                if counterfactual_config.is_computed and \
                        not counterfactual_config.has_computation_failed:
                    counterfactual_obj_list.append(
                        counterfactual_config.counterfactual_obj)
            return counterfactual_obj_list
        else:
            failure_reason_list = []
            for counterfactual_config in self._counterfactual_config_list:
                if counterfactual_config.is_computed and \
                        counterfactual_config.has_computation_failed:
                    failure_reason_list.append(
                        counterfactual_config.failure_reason)
            return failure_reason_list

    @staticmethod
    def _get_counterfactual_schema(version):
        """Get the schema for validating the counterfactual examples output."""
        schema_directory = (Path(__file__).parent.parent / '_tools' /
                            'counterfactual' / 'dashboard_schemas')
        schema_filename = f'counterfactual_examples_output_v{version}.json'
        schema_filepath = schema_directory / schema_filename
        with open(schema_filepath, 'r') as f:
            return json.load(f)

    def list(self):
        """List information about the CounterfactualManager.

        :return: A dictionary of properties.
        :rtype: dict
        """
        props = {ListProperties.MANAGER_TYPE: self.name}
        counterfactual_props_list = []
        for counterfactual_config in self._counterfactual_config_list:
            cf_coufig_dict = counterfactual_config.get_config_as_dict()
            cf_coufig_dict.update(counterfactual_config.get_result(
                include_counterfactual_obj=False))
            counterfactual_props_list.append(cf_coufig_dict)

        props[CounterfactualManagerKeys.COUNTERFACTUALS] = \
            counterfactual_props_list

        return props

    def get_data(self):
        """Get counterfactual data

        :return: A array of CounterfactualData.
        :rtype: List[CounterfactualData]
        """
        serialized_counterfactual_data_list = []
        for counterfactual_config in self._counterfactual_config_list:
            serialized_counterfactual_data_list.append(
                self._get_counterfactual(counterfactual_config))

        return serialized_counterfactual_data_list

    def _get_counterfactual(self, counterfactual_config,
                            counterfactual_object=None):
        cfdata = CounterfactualData()

        if counterfactual_object is None:
            json_data = json.loads(
                counterfactual_config.counterfactual_obj.to_json())
        else:
            json_data = json.loads(counterfactual_object.to_json())

        cfdata.cfs_list = json_data["cfs_list"]
        cfdata.feature_names = json_data["feature_names"]
        cfdata.feature_names_including_target = json_data[
            "feature_names_including_target"]
        cfdata.summary_importance = json_data["summary_importance"]
        cfdata.local_importance = json_data["local_importance"]
        cfdata.model_type = json_data["model_type"]
        cfdata.desired_class = json_data["desired_class"]
        cfdata.desired_range = json_data["desired_range"]
        cfdata.id = counterfactual_config.id
        cfdata.test_data = json_data["test_data"]
        return cfdata

    @property
    def name(self):
        """Get the name of the counterfactual manager.

        :return: The name of the counterfactual manager.
        :rtype: str
        """
        return ManagerNames.COUNTERFACTUAL

    def _save(self, path):
        """Save the CounterfactualManager to the given path.

        :param path: The directory path to save the CounterfactualManager to.
        :type path: str
        """
        counterfactual_dir = Path(path)
        counterfactual_dir.mkdir(parents=True, exist_ok=True)
        for counterfactual_config in self._counterfactual_config_list:
            directory_manager = DirectoryManager(parent_directory_path=path)

            counterfactual_config.save_config(
                directory_manager.create_config_directory()
            )

            counterfactual_config.save_result(
                directory_manager.create_data_directory()
            )

            counterfactual_config.save_explainer(
                directory_manager.create_generators_directory()
            )

    @staticmethod
    def _load(path, rai_insights):
        """Load the CounterfactualManager from the given path.

        :param path: The directory path to load the CounterfactualManager from.
        :type path: str
        :param rai_insights: The loaded parent RAIInsights.
        :type rai_insights: RAIInsights
        :return: The CounterfactualManager manager after loading.
        :rtype: CounterfactualManager
        """
        inst = CounterfactualManager.__new__(CounterfactualManager)

        # Rehydrate model analysis data
        inst.__dict__[CounterfactualManager._MODEL] = rai_insights.model
        inst.__dict__[CounterfactualManager._TRAIN] = rai_insights.train
        inst.__dict__[CounterfactualManager._TEST] = rai_insights.test
        inst.__dict__[CounterfactualManager._TARGET_COLUMN] = \
            rai_insights.target_column
        inst.__dict__[CounterfactualManager._TASK_TYPE] = \
            rai_insights.task_type
        inst.__dict__[CounterfactualManager._CATEGORICAL_FEATURES] = \
            rai_insights.categorical_features

        inst.__dict__[CounterfactualManager._COUNTERFACTUAL_CONFIG_LIST] = []

        # DirectoryManager.ensure_dir_exists(path)
        all_cf_dirs = DirectoryManager.list_sub_directories(path)
        for counterfactual_config_dir in all_cf_dirs:
            directory_manager = DirectoryManager(
                parent_directory_path=path,
                sub_directory_name=counterfactual_config_dir)

            counterfactual_config = CounterfactualConfig.load_config(
                directory_manager.get_config_directory()
            )

            counterfactual_config.load_result(
                directory_manager.get_data_directory()
            )

            counterfactual_config.load_explainer(
                directory_manager.get_generators_directory()
            )

            if counterfactual_config.explainer is None:
                explainer_load_err = (
                    'ERROR-LOADING-COUNTERFACTUAL-EXPLAINER: '
                    'There was an error loading the '
                    'counterfactual explainer model. '
                    'Retraining the counterfactual '
                    'explainer.')
                warnings.warn(explainer_load_err)
                counterfactual_config.explainer = \
                    inst._create_diceml_explainer(
                        counterfactual_config.method,
                        counterfactual_config.continuous_features)

            if counterfactual_config.counterfactual_obj is not None:
                # Validate the serialized output against schema
                schema = CounterfactualManager._get_counterfactual_schema(
                    version=counterfactual_config.counterfactual_obj.metadata[
                        'version'])
                jsonschema.validate(
                    json.loads(
                        counterfactual_config.counterfactual_obj.to_json()),
                    schema)

            inst.__dict__[
                CounterfactualManager._COUNTERFACTUAL_CONFIG_LIST].append(
                    counterfactual_config)

        return inst



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Data validations for responsibleai module."""
from typing import List, Optional

import numpy as np
import pandas as pd

from raiutils.exceptions import UserConfigValidationException


def _validate_unique_operation_on_categorical_columns(
        train_data: pd.DataFrame,
        test_data: pd.DataFrame,
        categorical_features: List[str]) -> None:
    """Validate unique operation on categorical columns.

    :param train_data: Training data.
    :type train_data: pd.DataFrame
    :param test_data: Test data.
    :type test_data: pd.DataFrame
    :param categorical_features: List of categorical features.
    :type categorical_features: List[str]
    :raises UserConfigValidationException: If unique operation is not
        successful on categorical columns.
    :return: None
    """
    for column in categorical_features:
        try:
            np.unique(train_data[column])
        except Exception:
            raise UserConfigValidationException(
                f"Error finding unique values in column {column}."
                " Please check your train data."
            )

        try:
            np.unique(test_data[column])
        except Exception:
            raise UserConfigValidationException(
                f"Error finding unique values in column {column}. "
                "Please check your test data.")


def validate_train_test_categories(
    train_data: pd.DataFrame,
    test_data: pd.DataFrame,
    rai_compute_type: str,
    categoricals: Optional[List[str]] = None,
):
    if categoricals is None:
        return
    _validate_unique_operation_on_categorical_columns(
        train_data, test_data, categoricals
    )
    discovered = {}
    for column in categoricals:
        if column in train_data.columns:
            train_unique = np.unique(train_data[column])
            test_unique = np.unique(test_data[column])
            difference = np.setdiff1d(test_unique, train_unique)
            if difference.shape[0] != 0:
                discovered[column] = difference.tolist()
    if len(discovered) > 0:
        message = ("{} requires that every category of "
                   "categorical features present in the test data "
                   "be also present in the train data. "
                   "Categories missing from train data: {}")
        raise UserConfigValidationException(
            message.format(rai_compute_type, discovered)
        )



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

from typing import Callable, Dict, List

import numpy as np
import pandas as pd
import scipy

from responsibleai.databalanceanalysis.balance_measures import BalanceMeasures
from responsibleai.databalanceanalysis.constants import Constants, Measures


def get_cross_entropy(obs: np.array, ref: np.array) -> float:
    """
    Returns the cross entropy between two probability distributions.
    https://en.wikipedia.org/wiki/Cross_entropy

    :param obs: The observed probability distribution.
    :type obs: np.array
    :param ref: The reference probability distribution.
    :type ref: np.array
    :return: The cross entropy between the two probability distributions.
    :rtype: float
    """
    return -np.sum(ref * np.log2(obs))


def get_kl_divergence(obs: np.array, ref: np.array) -> float:
    """
    Returns the Kullback-Leibler (KL) divergence, a measure of how one
    probability distribution is different from another.
    https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence

    :param obs: The observed probability distribution.
    :type obs: np.array
    :param ref: The reference probability distribution.
    :type ref: np.array
    :return: The KL divergence between the two probability distributions.
    :rtype: float
    """
    return scipy.stats.entropy(obs, qk=ref)


def get_js_distance(obs: np.array, ref: np.array) -> float:
    """
    Returns the Jenson-Shannon Distance, a measure that is the
    symmetrized and smoothed version of the KL divergence.
    https://en.wikipedia.org/wiki/Jensen%E2%80%93Shannon_divergence

    :param obs: The observed probability distribution.
    :type obs: np.array
    :param ref: The reference probability distribution.
    :type ref: np.array
    :return: The JS Distance between the two probability distributions.
    :rtype: float
    """
    avg = (obs + ref) / 2
    divergence = (
        scipy.stats.entropy(obs, avg) + scipy.stats.entropy(ref, avg)
    ) / 2
    distance = np.sqrt(divergence)
    return distance


def get_ws_distance(obs: np.array, ref: np.array) -> float:
    """
    Returns the Wasserstein Distance, a measure that symbolizes the
    minimum cost to turn one probability distribution into another.
    https://en.wikipedia.org/wiki/Wasserstein_metric

    :param obs: The observed probability distribution.
    :type obs: np.array
    :param ref: The reference probability distribution.
    :type ref: np.array
    :return: The Wasserstein Distance between the two probability
        distributions.
    :rtype: float
    """
    return scipy.stats.wasserstein_distance(obs, ref)


def get_inf_norm_dist(obs: np.array, ref: np.array) -> float:
    """
    Returns the Infinity norm distance (AKA the Chebyshev distance),
    a measure of the maximum absolute difference between two probability
    distributions.
    https://en.wikipedia.org/wiki/Chebyshev_distance

    :param obs: The observed probability distribution.
    :type obs: np.array
    :param ref: The reference probability distribution.
    :type ref: np.array
    :return: The Infinity norm distance between the two probability
        distributions.
    :rtype: float
    """
    return scipy.spatial.distance.chebyshev(obs, ref)


def get_total_var_dist(obs: np.array, ref: np.array) -> float:
    """
    Returns the total variation distance, a measure similar to the KL
    divergence in specifying the distance between two probability
    distributions.
    https://en.wikipedia.org/wiki/Total_variation_distance_of_probability_measures

    :param obs: The observed probability distribution.
    :type obs: np.array
    :param ref: The reference probability distribution.
    :type ref: np.array
    :return: The total variation distance between the two probability
        distributions.
    :rtype: float
    """
    return 0.5 * np.sum(np.abs(obs - ref))


def get_chisq_statistic(f_obs: np.array, f_ref: np.array) -> float:
    """
    Returns the Chi-Square test statistic, a measure that determines the
    statistical significant difference between two frequencies.
    https://en.wikipedia.org/wiki/Chi-squared_test

    :param f_obs: Frequencies of the observed probability distribution.
    :type f_obs: np.array
    :param f_ref: Frequencies of the reference probability distribution.
    :type f_ref: np.array
    :return: The Chi-Square test statistic between the frequencies of the
        two probability distributions.
    :rtype: float
    """
    res = scipy.stats.chisquare(f_obs, f_ref)
    chisq: float = res[0]
    return chisq


def get_chisq_pvalue(f_obs: np.array, f_ref: np.array) -> float:
    """
    Returns the p-value of the Chi-Square, which is the probability of
    obtaining test results at least as extreme as the observed results.
    A small p-value means that the observed results are very unlikely
    under the null hypothesis (i.e. the distributions are the same).
    https://en.wikipedia.org/wiki/Chi-squared_test

    :param f_obs: Frequencies of the observed probability distribution.
    :type f_obs: np.array
    :param f_ref: Frequencies of the reference probability distribution.
    :type f_ref: np.array
    :return: The p-value of the Chi-Square test statistic between the
        frequencies of the two probability distributions.
    :rtype: float
    """
    res = scipy.stats.chisquare(f_obs, f_ref)
    pvalue: float = res[1]
    return pvalue


class DistributionBalanceMeasures(BalanceMeasures):
    """
    This class computes data balance measures for columns of interest
    based on a reference distribution. Currently, only the uniform reference
    distribution is supported.
    """

    DISTRIBUTION_METRICS: Dict[
        Measures, Callable[[np.array, np.array], float]
    ] = {
        Measures.KL_DIVERGENCE: get_kl_divergence,
        Measures.JS_DISTANCE: get_js_distance,
        Measures.WS_DISTANCE: get_ws_distance,
        Measures.INF_NORM_DIST: get_inf_norm_dist,
        Measures.TOTAL_VAR_DIST: get_total_var_dist,
        Measures.CHISQ_PVALUE: get_chisq_pvalue,
        Measures.CHISQ_STATISTIC: get_chisq_statistic,
        Measures.CROSS_ENTROPY: get_cross_entropy,
    }

    def __init__(self, cols_of_interest: List[str]) -> None:
        """
        Construct a DistributionBalanceMeasure class for calculating
        distribution balance measures on a dataset.

        :param cols_of_interest: The list of columns to compute
            distribution balance measures on.
        :type cols_of_interest: List[str]
        """
        super().__init__(cols_of_interest=cols_of_interest)

    def measures(self, dataset: pd.DataFrame) -> pd.DataFrame:
        """
        Returns distribution balance measures for the given dataset and
        columns of interest.

        The following measures are computed:
        * Kullback-Leibler Divergence -
            https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence
        * Jensen-Shannon Distance -
            https://en.wikipedia.org/wiki/Jensen%E2%80%93Shannon_divergence
        * Wasserstein Distance -
            https://en.wikipedia.org/wiki/Wasserstein_metric
        * Infinity Norm Distance -
            https://en.wikipedia.org/wiki/Chebyshev_distance
        * Total Variation Distance -
            https://en.wikipedia.org/wiki/Total_variation_distance_of_probability_measures
        * Chi-Squared Test -
            https://en.wikipedia.org/wiki/Chi-squared_test
        * Cross Entropy -
            https://en.wikipedia.org/wiki/Cross_entropy

        :param dataset: The dataset to compute distribution measures on.
        :type dataset: pd.DataFrame
        :return: A pandas DataFrame that contains the given
            columns of interest and for each column of interest,
            a `dict` of <measure name, measure value> pairs.
        :rtype: pd.DataFrame
        """
        all_measures = [
            DistributionBalanceMeasures._get_distribution_measures_for_col(
                df=dataset, col_of_interest=col
            )
            for col in self.cols_of_interest
        ]
        return pd.DataFrame.from_dict(all_measures)

    @staticmethod
    def _get_distribution_measures_for_col(
        df: pd.DataFrame, col_of_interest: str
    ) -> Dict[Measures, float]:
        """
        Returns a dictionary of distribution balance measures based on the
        given dataset and column of interest.

        :param df: The dataset to compute distribution measures on.
        :type df: pd.DataFrame
        :param col_of_interest: The column of interest to compute
            distribution measures on.
        :type col_of_interest: str
        :return: A `dict` of <measure name, measure value> pairs.
        :rtype: Dict[Measures, float]
        """
        f_obs = df.groupby(col_of_interest).size().reset_index(name="count")
        sum_obs = f_obs["count"].sum()
        obs = f_obs["count"] / sum_obs
        ref = DistributionBalanceMeasures._create_reference_distribution(
            n=f_obs.shape[0]
        )
        f_ref = ref * sum_obs

        measures = {Constants.FEATURE_NAME.value: col_of_interest}
        for (
            measure,
            func,
        ) in DistributionBalanceMeasures.DISTRIBUTION_METRICS.items():
            if measure in [Measures.CHISQ_PVALUE, Measures.CHISQ_STATISTIC]:
                measures[measure.value] = func(f_obs["count"], f_ref)
            else:
                measures[measure.value] = func(obs, ref)

        return measures

    @staticmethod
    def _create_reference_distribution(n: int) -> np.array:
        """
        Returns a reference distribution based on `n` outcomes.
        For now, this is a uniform distribution.

        :param n: The number of outcomes.
        :type n: int
        :return: The reference distribution.
        :rtype: numpy.ndarray
        """
        uniform_val: float = 1.0 / n
        return np.ones(n) * uniform_val



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Note: this test file will be removed once ModelAnalysis is removed."""

import os
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd
import pytest
from tests.causal_manager_validator import validate_causal
from tests.common_utils import create_iris_data
from tests.counterfactual_manager_validator import validate_counterfactual
from tests.error_analysis_validator import (setup_error_analysis,
                                            validate_error_analysis)
from tests.explainer_manager_validator import (setup_explainer,
                                               validate_explainer)

from rai_test_utils.datasets.tabular import (
    create_binary_classification_dataset, create_cancer_data,
    create_housing_data)
from rai_test_utils.models.model_utils import (create_models_classification,
                                               create_models_regression)
from rai_test_utils.models.sklearn import \
    create_complex_classification_pipeline
from rai_test_utils.utilities import is_valid_uuid
from raiutils.models import ModelTask
from responsibleai import ModelAnalysis
from responsibleai._internal.constants import ManagerNames
from responsibleai._tools.shared.state_directory_management import \
    DirectoryManager

LABELS = 'labels'


class ManagerParams:
    # Counterfactual
    DESIRED_CLASS = 'desired_class'
    DESIRED_RANGE = 'desired_range'
    FEATURE_IMPORTANCE = 'feature_importance'

    # Causal
    TREATMENT_FEATURES = 'treatment_features'
    MAX_CAT_EXPANSION = 'max_cat_expansion'


class TestModelAnalysis(object):

    @pytest.mark.parametrize('manager_type', [ManagerNames.COUNTERFACTUAL,
                                              ManagerNames.ERROR_ANALYSIS,
                                              ManagerNames.CAUSAL,
                                              ManagerNames.EXPLAINER])
    def test_model_analysis_iris(self, manager_type):
        X_train, X_test, y_train, y_test, feature_names, classes = \
            create_iris_data()
        models = create_models_classification(X_train, y_train)
        X_train[LABELS] = y_train
        X_test[LABELS] = y_test
        manager_args = {
            ManagerParams.TREATMENT_FEATURES: [feature_names[0]],
            ManagerParams.DESIRED_CLASS: 0,
            ManagerParams.FEATURE_IMPORTANCE: True
        }

        for model in models:
            run_model_analysis(model, X_train, X_test, LABELS, None,
                               manager_type, manager_args, classes)

    @pytest.mark.parametrize('manager_type', [ManagerNames.ERROR_ANALYSIS,
                                              ManagerNames.COUNTERFACTUAL,
                                              ManagerNames.EXPLAINER])
    def test_model_analysis_cancer(self, manager_type):
        X_train, X_test, y_train, y_test, _, classes = \
            create_cancer_data(return_dataframe=True)
        models = create_models_classification(X_train, y_train)
        X_train[LABELS] = y_train
        X_test[LABELS] = y_test
        manager_args = {
            ManagerParams.DESIRED_CLASS: 'opposite',
            ManagerParams.FEATURE_IMPORTANCE: False
        }

        for model in models:
            run_model_analysis(model, X_train, X_test, LABELS, None,
                               manager_type, manager_args, classes)

    @pytest.mark.parametrize('manager_type', [ManagerNames.CAUSAL,
                                              ManagerNames.ERROR_ANALYSIS,
                                              ManagerNames.EXPLAINER])
    def test_model_analysis_binary(self, manager_type):
        X_train, y_train, X_test, y_test, classes = \
            create_binary_classification_dataset()

        models = create_models_classification(X_train, y_train)
        X_train[LABELS] = y_train
        X_test[LABELS] = y_test
        manager_args = {
            ManagerParams.TREATMENT_FEATURES: ['col0']
        }

        for model in models:
            run_model_analysis(model, X_train, X_test, LABELS, None,
                               manager_type, manager_args,
                               classes=classes)

    @pytest.mark.parametrize('manager_type', [ManagerNames.CAUSAL,
                                              ManagerNames.ERROR_ANALYSIS,
                                              ManagerNames.EXPLAINER,
                                              ManagerNames.COUNTERFACTUAL])
    def test_model_analysis_binary_mixed_types(self, manager_type, adult_data):

        data_train, data_test, y_train, y_test, categorical_features, \
            continuous_features, target_name, classes, \
            feature_columns, feature_range_keys = adult_data
        X_train = data_train.drop([target_name], axis=1)

        model = create_complex_classification_pipeline(
            X_train, y_train, continuous_features, categorical_features)
        manager_args = {
            ManagerParams.TREATMENT_FEATURES: ['age', 'hours_per_week'],
            ManagerParams.DESIRED_CLASS: 'opposite',
            ManagerParams.FEATURE_IMPORTANCE: False
        }

        run_model_analysis(model, data_train, data_test, target_name,
                           categorical_features,
                           manager_type, manager_args,
                           classes=classes)

    @pytest.mark.parametrize('manager_type', [ManagerNames.CAUSAL,
                                              ManagerNames.ERROR_ANALYSIS,
                                              ManagerNames.EXPLAINER,
                                              ManagerNames.COUNTERFACTUAL])
    def test_model_analysis_no_model(self, manager_type):

        X_train, y_train, X_test, y_test, classes = \
            create_binary_classification_dataset()

        X_train[LABELS] = y_train
        X_test[LABELS] = y_test
        model = None
        manager_args = {
            ManagerParams.TREATMENT_FEATURES: ['col0']
        }

        run_model_analysis(model, X_train, X_test, LABELS, None,
                           manager_type, manager_args,
                           classes=classes)

    @pytest.mark.parametrize('manager_type', [ManagerNames.CAUSAL,
                                              ManagerNames.COUNTERFACTUAL,
                                              ManagerNames.EXPLAINER])
    def test_modelanalysis_housing(self, manager_type):
        X_train, X_test, y_train, y_test, feature_names = \
            create_housing_data()
        X_train = pd.DataFrame(X_train, columns=feature_names)
        X_test = pd.DataFrame(X_test, columns=feature_names)
        models = create_models_regression(X_train, y_train)
        X_train[LABELS] = y_train
        X_test[LABELS] = y_test

        manager_args = {
            ManagerParams.DESIRED_RANGE: [3, 5],
            ManagerParams.TREATMENT_FEATURES: ['AveRooms'],
            ManagerParams.MAX_CAT_EXPANSION: 12,
            ManagerParams.FEATURE_IMPORTANCE: True
        }

        for model in models:
            run_model_analysis(model, X_train, X_test, LABELS, [],
                               manager_type, manager_args)


def run_model_analysis(model, train_data, test_data, target_column,
                       categorical_features, manager_type,
                       manager_args=None, classes=None):
    if manager_args is None:
        manager_args = {}

    if classes is not None:
        task_type = ModelTask.CLASSIFICATION
    else:
        task_type = ModelTask.REGRESSION

    if manager_type == ManagerNames.COUNTERFACTUAL:
        feature_importance = manager_args.get(ManagerParams.FEATURE_IMPORTANCE)
        if feature_importance:
            test_data = test_data[0:20]
        else:
            test_data = test_data[0:1]

    if model is None:
        with pytest.warns(
                UserWarning,
                match='INVALID-MODEL-WARNING: '
                      'No valid model is supplied. '
                      'The explanations, error analysis and '
                      'counterfactuals may not work'):
            model_analysis = ModelAnalysis(
                model, train_data, test_data,
                target_column,
                categorical_features=categorical_features,
                task_type=task_type)
    else:
        with pytest.warns(DeprecationWarning,
                          match=("MODULE-DEPRECATION-WARNING: "
                                 "ModelAnalysis in responsibleai "
                                 "package is deprecated. "
                                 "Please use RAIInsights instead.")):
            model_analysis = ModelAnalysis(
                model, train_data, test_data,
                target_column,
                categorical_features=categorical_features,
                task_type=task_type)

    if manager_type == ManagerNames.EXPLAINER:
        setup_explainer(model_analysis)
    elif manager_type == ManagerNames.ERROR_ANALYSIS:
        setup_error_analysis(model_analysis)

    validate_model_analysis(model_analysis, train_data, test_data,
                            target_column, task_type, categorical_features)

    if manager_type == ManagerNames.CAUSAL:
        treatment_features = manager_args.get(ManagerParams.TREATMENT_FEATURES)
        max_cat_expansion = manager_args.get(ManagerParams.MAX_CAT_EXPANSION)
        validate_causal(model_analysis, train_data, target_column,
                        treatment_features, max_cat_expansion)
    elif manager_type == ManagerNames.COUNTERFACTUAL:
        desired_class = manager_args.get(ManagerParams.DESIRED_CLASS)
        desired_range = manager_args.get(ManagerParams.DESIRED_RANGE)
        feature_importance = manager_args.get(ManagerParams.FEATURE_IMPORTANCE)
        validate_counterfactual(model_analysis,
                                desired_class=desired_class,
                                desired_range=desired_range,
                                feature_importance=feature_importance)
    elif manager_type == ManagerNames.ERROR_ANALYSIS:
        validate_error_analysis(model_analysis)
    elif manager_type == ManagerNames.EXPLAINER:
        validate_explainer(model_analysis, train_data, test_data, classes)

    with TemporaryDirectory() as tempdir:
        path = Path(tempdir) / 'rai_test_path'

        # save the model_analysis
        model_analysis.save(path)

        # Validate the directory structure of the state saved
        # by the managers.
        validate_state_directory(path, manager_type)

        # load the model_analysis
        model_analysis = ModelAnalysis.load(path)

        validate_model_analysis(
            model_analysis, train_data, test_data,
            target_column, task_type, categorical_features)

        if manager_type == ManagerNames.ERROR_ANALYSIS:
            validate_error_analysis(model_analysis)
            # validate adding new reports after deserialization works
            setup_error_analysis(model_analysis, max_depth=4)
            validate_error_analysis(model_analysis, expected_reports=2)
        elif manager_type == ManagerNames.EXPLAINER:
            validate_explainer(model_analysis, train_data, test_data, classes)
            # validate adding new explainer config after deserialization works
            setup_explainer(model_analysis)
            validate_explainer(model_analysis, train_data, test_data, classes)


def validate_state_directory(path, manager_type):
    all_dirs = os.listdir(path)
    assert manager_type in all_dirs
    all_component_paths = os.listdir(path / manager_type)
    for component_path in all_component_paths:
        # Test if the component directory has UUID structure
        is_valid_uuid(component_path)

        dm = DirectoryManager(path / manager_type, component_path)

        config_path = dm.get_config_directory()
        data_path = dm.get_data_directory()
        generators_path = dm.get_generators_directory()

        if manager_type == ManagerNames.EXPLAINER:
            assert not config_path.exists()
            assert data_path.exists()
            assert not generators_path.exists()
        elif manager_type == ManagerNames.COUNTERFACTUAL:
            assert config_path.exists()
            assert data_path.exists()
            assert generators_path.exists()
        elif manager_type == ManagerNames.ERROR_ANALYSIS:
            assert config_path.exists()
            assert data_path.exists()
            assert not generators_path.exists()
        elif manager_type == ManagerNames.CAUSAL:
            assert not config_path.exists()
            assert data_path.exists()
            assert not generators_path.exists()


def validate_model_analysis(
    model_analysis,
    train_data,
    test_data,
    target_column,
    task_type,
    categorical_features
):

    pd.testing.assert_frame_equal(model_analysis.train, train_data)
    pd.testing.assert_frame_equal(model_analysis.test, test_data)
    assert model_analysis.target_column == target_column
    assert model_analysis.task_type == task_type
    assert model_analysis.categorical_features == (categorical_features or [])
    assert type(model_analysis.categorical_features) is list
    if task_type == ModelTask.CLASSIFICATION:
        classes = train_data[target_column].unique()
        classes.sort()
        np.testing.assert_array_equal(model_analysis.rai_insights._classes,
                                      classes)
